import beartype.typing as tp
import jax.numpy as jnp
from beartype.typing import Literal
from jax import jit, lax
from jax import random as jr
from jaxtyping import Int, PRNGKeyArray

from gallifrey.utils.typing import ScalarBool, ScalarFloat, ScalarInt


def get_child_idx(idx: ScalarInt, direction: Literal["l", "r"]) -> ScalarInt:
    """
    Get the index of the left or right child of a node in a binary tree,
    given by
    2 * idx + 1 for left child
    2 * idx + 2 for right child.

    Compare: https://en.wikipedia.org/wiki/Binary_tree#Arrays

    Parameters
    ----------
    idx : ScalarInt
        The index of the node.
    direction : Literal["l", "r"]
        The side of the child to get. "l" for left child, "r" for right child.

    Returns
    -------
    ScalarInt
        The index of the child.
    """
    direction_map = {"l": 1, "r": 2}
    return 2 * idx + direction_map[direction]


def get_parent_idx(idx: ScalarInt) -> ScalarInt:
    """
    Get the index of the parent of a node in a binary tree,
    given by floor((idx - 1) / 2).

    Compare: https://en.wikipedia.org/wiki/Binary_tree#Arrays

    Parameters
    ----------
    idx : ScalarInt
        The index of the node.

    Returns
    -------
    int
        The index of the parent.
    """
    return (idx - 1) // 2


def get_depth(idx: ScalarInt, tolerance: ScalarFloat = 1e-10) -> ScalarInt:
    """
    Get the depth of a node in a binary tree.

    Parameters
    ----------
    idx : ScalarInt
        The index of the node.
    tolerance : ScalarFloat, optional
        On occasion, floating point errors may lead to a depth that is off by one.
        This can happen whenever the index is a power of 2. To avoid this, a tolerance
        is added to the index before taking the logarithm. By default 1e-10.

    Returns
    -------
    int
        The depth of the node.
    """
    return jnp.log2(idx + 1 + tolerance).astype(int)


def calculate_max_nodes(max_depth: ScalarInt) -> ScalarInt:
    """
    Calculate the maximum number of nodes in a binary tree based
    on the maximum depth of the tree.
     Since each node in the tree can have at most two children,
    the maximum number of nodes in the tree is given by
    2^(max_depth + 1) - 1.

    Parameters
    ----------
    max_depth : ScalarInt
        The maximum depth of the nested kernel structure tree.

    Returns
    -------
    ScalarInt
        The maximum number of nodes in the tree that describes the kernel structure.
    """
    return 2 ** (max_depth + 1) - 1


def calculate_max_stack_size(max_depth: ScalarInt) -> ScalarInt:
    """
    For a given tree depth, calculate the maximum size of the stack
    needed to traverse the tree in level order.

    Parameters
    ----------
    max_depth : ScalarInt
        The maximum depth of the nested kernel structure tree.

    Returns
    -------
    ScalarInt
        The maximum size of the stack needed to traverse the tree.
    """
    return max_depth + 1


def calculate_max_leaves(max_depth: ScalarInt) -> ScalarInt:
    """
    Calculate the maximum number of leaves in a binary tree based
    on the maximum depth of the tree.
    This corresponds to the number of nodes at the maximum depth
    of the tree, which is given by 2^max_depth.

    Parameters
    ----------
    max_depth : ScalarInt
        The maximum depth of the nested kernel structure tree.

    Returns
    -------
    ScalarInt
        The maximum number of leaves in the tree that describes the
        kernel structure.
    """
    return 2**max_depth


def calculate_max_depth(
    max_nodes: ScalarInt,
    tolerance: ScalarFloat = 1e-10,
) -> ScalarInt:
    """
    Calculate the maximum depth of the tree that describes the kernel structure,
    given the maximum number of nodes in the tree.

    Parameters
    ----------
    max_nodes : ScalarInt
        The maximum number of nodes in the tree that describes the kernel structure.
    tolerance : ScalarFloat, optional
        On occasion, floating point errors may lead to a depth that is off by one.
        This can happen whenever the index is a power of 2. To avoid this, a tolerance
        is added to the index before taking the logarithm. By default 1e-10.

    Returns
    -------
    ScalarInt
        The maximum depth of the nested kernel structure tree.
    """
    return get_depth(max_nodes - 1, tolerance)  # idx of last node is max_nodes - 1


def get_leftmost_leaf(idx: ScalarInt, max_depth: ScalarInt) -> ScalarInt:
    """
    For a tree with a given maximum depth, get the index of the leftmost child
    leaf at the maximum depth from a given index (in level order notation).

    Since the equation for a left child is 2 * idx + 1, the left child of the left
    child is 4*i+3, and in general the left-most child at max depth d is
    2^(max_depth-current_depth)*idx + 2^(max_depth-current_depth)-1 =
    2^(max_depth-current_depth)*(idx+1) - 1., which
    can be efficiently implemented using bit-shift operations.


    Parameters
    ----------
    idx : _type_
        _description_
    max_depth : _type_
        _description_

    Returns
    -------
    ScalarInt
        The index of the leftmost leaf at the maximum depth.
    """
    current_depth = get_depth(idx)
    diff = max_depth - current_depth
    return jnp.left_shift(idx + 1, diff) - 1


def get_parameter_leaf_idx(
    tree_level_idx: ScalarInt,
    max_depth: ScalarInt,
) -> ScalarInt:
    """
    Get the index in the parameter array for a given node in the tree expression.

    We work with the convention that, if a leaf node is not at the maximum depth,
    the parameters are stored in the leftmost leaf at the maximum depth. Substracting
    the index of the very left most leaf (from the root) therefore gives a unique index
    for each possible leaf node in the tree.

    Parameters
    ----------
    tree_level_idx : ScalarInt
        The index of the node in the tree expression (level order).
    max_depth : ScalarInt
        The maximum depth of the tree.

    Returns
    -------
    ScalarInt
        The index of the parameters in the parameter array for a given node.

    """
    return get_leftmost_leaf(tree_level_idx, max_depth) - get_leftmost_leaf(
        0, max_depth
    )


def clear_subtree(
    tree_expression: Int[jnp.ndarray, " D"],
    clear_idx: ScalarInt,
    fill_value: ScalarFloat = -1,
) -> Int[jnp.ndarray, " D"]:
    """
    Clear a subtree in a level order tree expression.

    NOTE: The new tree expression will not be a valid
    kernel expression, as the subtree removal leaves
    a hole in the expression.

    Parameters
    ----------
    tree_expression : Int[jnp.ndarray, " D"]
        The tree expression, given in level order notation.
    clear_idx : ScalarInt
        The index of the root node of the subtree to clear.
    fill_value : ScalarFloat, optional
        The value to fill the cleared nodes with. By default
        -1.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The updated tree expression with the subtree cleared.

    """
    max_nodes = tree_expression.size

    # create initial state
    stack = jnp.zeros(max_nodes, dtype=tree_expression.dtype)
    stack = stack.at[0].set(clear_idx)
    stack_pointer = 0
    initial_state = (tree_expression, stack, stack_pointer)

    return _clear_subtree(
        initial_state,
        max_nodes,
        fill_value,
    )


@jit
def _clear_subtree(
    initial_state: tuple[
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
        ScalarInt,
    ],
    max_nodes: ScalarInt,
    fill_value: ScalarInt,
) -> Int[jnp.ndarray, " D"]:
    """
    Clear a subtree in a level order tree expression.

    NOTE: This function is not meant to be called directly. Use clear_subtree
    instead.

    Parameters
    ----------
    initial_state : tuple
        The initial state of the traversal. Contains the tree expression,
        the stack and the stack pointer.
    max_nodes : ScalarInt
        The maximum number of nodes in the tree.
    fill_value : ScalarInt
        The value to fill the cleared nodes with.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The updated tree expression with the subtree cleared.

    """

    def traverse(state: tuple) -> tuple:
        """Traverse the tree expression in level-order."""
        tree_expression, stack, pointer = state

        # clear the current node
        idx = stack[pointer]
        tree_expression_updated = tree_expression.at[idx].set(fill_value)

        # check if the current node is a leaf
        right_child_idx = get_child_idx(idx, "r")
        is_leaf = jnp.logical_or(
            right_child_idx > max_nodes - 1,
            tree_expression[right_child_idx - 1] == -1,
        )

        def process_leaf(stack_and_pointer: tuple) -> tuple:
            """If the current node is a leaf, move to the parent node
            (node has already been cleared)."""
            stack, pointer = stack_and_pointer
            return stack, pointer - 1

        def process_non_leaf(stack_and_pointer: tuple) -> tuple:
            """If the current node is not a leaf, update the stack
            (add the children of the current node)."""
            stack, pointer = stack_and_pointer

            left_child_idx = get_child_idx(idx, "l")
            right_child_idx = get_child_idx(idx, "r")

            new_stack = stack.at[pointer].set(left_child_idx)
            new_stack = new_stack.at[pointer + 1].set(right_child_idx)

            return new_stack, pointer + 1

        # update the stack and pointer
        new_stack, new_pointer = lax.cond(
            is_leaf,
            process_leaf,
            process_non_leaf,
            operand=(stack, pointer),
        )

        return tree_expression_updated, new_stack, new_pointer

    def condition(state: tuple) -> ScalarBool:
        """Check if the traversal is complete."""
        _, _, pointer = state
        return pointer >= 0

    # iterate over the sample using lax.while_loop
    new_tree_expression, _, _ = lax.while_loop(
        condition,
        traverse,
        initial_state,
    )
    return new_tree_expression


def get_subtree(
    tree_expression: Int[jnp.ndarray, " D"],
    subtree_root_idx: ScalarInt,
    new_root_idx: tp.Optional[ScalarInt] = None,
) -> tuple[
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, "2 D"],
]:
    """
    Retrieve a subtree from a level order tree expression, starting
    at a given attachment node.

    The tree can be re-rooted at a different node, if wanted.

    Parameters
    ----------
    tree_expression : Int[jnp.ndarray, " D"]
        The tree expression, given in level order notation.
    subtree_root_idx : ScalarInt
        The index of node where the subtree is rooted.
    new_root_idx : tp.Optional[ScalarInt], optional
        The new index of the root node, if the subtree is to be re-rooted,
        by default None. If None, the subtree is returned in its original
        location.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The (level order) tree expression of the subtree, same length as the
        input tree expression.
    Int[jnp.ndarray, "2 D"]
        The indices of the nodes in the subtree. Two rows, the first row
        contains the indices in the original tree, the second row contains
        the indices in the subtree. (Will be the same if the tree is not
        re-rooted.)

    """
    max_nodes = tree_expression.size

    # if tree_expression[subtree_root_idx] == -1:
    #     raise ValueError("The start index must be a non-empty node.")

    new_root_idx = subtree_root_idx if new_root_idx is None else new_root_idx

    # create initial state
    new_tree_expression = jnp.full(max_nodes, -1)
    index_array = jnp.full((2, max_nodes), -1)
    index_pointer = 0
    stack = jnp.copy(new_tree_expression).at[0].set(subtree_root_idx)
    new_stack = jnp.copy(stack).at[0].set(new_root_idx)
    stack_pointer = 0

    initial_state = (
        new_tree_expression,
        index_array,
        index_pointer,
        stack,
        new_stack,
        stack_pointer,
    )

    subtree_expression, index_array, _ = _get_subtree(
        tree_expression,
        initial_state,
        max_nodes,
    )

    index_array = index_array
    # if jnp.any(index_array >= max_nodes):
    #     raise ValueError(
    #         "The index array is out of bounds. This could happen if the tree "
    #         "is re-rooted to a spot where one of its leaves exceeds the "
    #         "maximum number of nodes."
    #     )
    return subtree_expression, index_array


@jit
def _get_subtree(
    tree_expression: Int[jnp.ndarray, " D"],
    initial_state: tuple[
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, "2 D"],
        ScalarInt,
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
        ScalarInt,
    ],
    max_nodes: ScalarInt,
) -> tuple[
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, "2 D"],
    ScalarInt,
]:
    """
    Traverse the tree expression in level-order, starting at a given node
    and retrieve the subtree rooted at that node. Perform a re-rooting if
    wanted.

    NOTE: This function is not meant to be called directly. Use get_subtree
    instead.

    Parameters
    ----------
    initial_state : tuple
        The initial state of the traversal. Contains the new tree expression,
        the new root index, the index array, the index pointer, the stack, the
        stack pointer, the new stack, and the new stack pointer.
    max_nodes : ScalarInt
        The maximum number of nodes in the tree.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The (level order) tree expression of the subtree, same length as the
        input tree expression.
    Int[jnp.ndarray, "2 D"]
        The indices of the nodes in the subtree. Two rows, the first row
        contains the indices in the original tree, the second row contains
        the indices in the subtree. (Will be the same if the tree is not
        re-rooted.)

    """

    def traverse(state: tuple) -> tuple:
        """Traverse the tree expression in level-order."""
        (
            subtree_expression,
            index_array,
            index_pointer,
            old_stack,
            new_stack,
            stack_pointer,
        ) = state

        # get old and new indices
        old_idx = old_stack[stack_pointer]
        new_idx = new_stack[stack_pointer]

        # fill index array
        index_array_updated = index_array.at[0, index_pointer].set(old_idx)
        index_array_updated = index_array_updated.at[1, index_pointer].set(new_idx)
        index_pointer_updated = index_pointer + 1

        # fill the new tree expression with the current node (at the new index)
        subtree_expression_updated = subtree_expression.at[new_idx].set(
            tree_expression[old_idx]
        )

        # check if the current node is a leaf
        right_child_idx = get_child_idx(old_idx, "r")
        is_leaf = jnp.logical_or(
            right_child_idx > max_nodes - 1,
            tree_expression[right_child_idx - 1] == -1,
        )

        def process_leaf(stacks_and_pointer: tuple) -> tuple:
            """If the current node is a leaf, move to the parent node."""
            old_stack, new_stack, pointer = stacks_and_pointer
            return old_stack, new_stack, pointer - 1

        def process_non_leaf(stacks_and_pointer: tuple) -> tuple:
            """If the current node is not a leaf, update the stacks
            (add the children of the current node)."""
            old_stack, new_stack, pointer = stacks_and_pointer

            left_child_idx = get_child_idx(old_idx, "l")
            right_child_idx = get_child_idx(old_idx, "r")

            new_left_child_idx = get_child_idx(new_idx, "l")
            new_right_child_idx = get_child_idx(new_idx, "r")

            old_stack_updated = old_stack.at[pointer].set(left_child_idx)
            old_stack_updated = old_stack_updated.at[pointer + 1].set(right_child_idx)

            new_stack_updated = new_stack.at[pointer].set(new_left_child_idx)
            new_stack_updated = new_stack_updated.at[pointer + 1].set(
                new_right_child_idx
            )

            return old_stack_updated, new_stack_updated, pointer + 1

        # update the stack and pointer
        old_stack_updated, new_stack_updated, pointer_updated = lax.cond(
            is_leaf,
            process_leaf,
            process_non_leaf,
            operand=(old_stack, new_stack, stack_pointer),
        )

        return (
            subtree_expression_updated,
            index_array_updated,
            index_pointer_updated,
            old_stack_updated,
            new_stack_updated,
            pointer_updated,
        )

    def condition(state: tuple) -> ScalarBool:
        """Check if the traversal is complete."""
        _, _, _, _, _, stack_pointer = state
        return stack_pointer >= 0

    # iterate over the sample using lax.while_loop
    final_subtree_expression, final_index_array, final_index_pointer, _, _, _ = (
        lax.while_loop(
            condition,
            traverse,
            initial_state,
        )
    )
    return final_subtree_expression, final_index_array, final_index_pointer


@jit
def pick_random_node(
    rng_key: PRNGKeyArray,
    post_level_map: Int[jnp.ndarray, " D"],
) -> ScalarInt:
    """
    Randomly pick a node from the post-order expression, return the
    corresponding node in level order. In the current implementation
    this is done by sampling uniformly from the non-empty nodes.

    The post_level_map contains all indices used in the the level
    order expression, which makes it extremely easy to sample a node
    uniformly at random. It is assumed at empty nodes are represented
    by -1 in the post_level_map, while all other nodes are have
    values >= 0.

    Parameters
    ----------
    rng_key : PRNGKeyArray
        Random key.
    post_level_map :  Int[jnp.ndarray, " D"]
        The map from post-order to level-order.

    Returns
    -------
    ScalarInt
        The randomly chosen node (in level order notation).

    """
    # set weights, mask empty nodes
    weights = jnp.where(post_level_map >= 0, 1.0, 0.0)
    return jr.choice(rng_key, post_level_map, p=weights)


def generate_random_path(
    key: PRNGKeyArray,
    start_idx: ScalarInt,
    node_height: ScalarInt,
    max_depth: ScalarInt,
    stop_probability: ScalarFloat = 0.5,
    go_left_probability: ScalarFloat = 0.5,
) -> tuple[
    ScalarInt,
    Int[jnp.ndarray, " D"],
    ScalarFloat,
]:
    """
    Generate a random path in a binary tree starting from a given index.
    Returns the final chosen index, the path taken and the log probability
    of reaching the chosen index.

    Parameters
    ----------
    key : jr.PRNGKey
        The random key.
    start_idx : ScalarInt
        The index of the starting node.
    max_depth : ScalarInt
        The maximum depth of the tree.
    node_height : ScalarInt
        The height of the subtree at the starting index. This is used
        to calculate the maximum number of levels the path can traverse,
        and still respect the maximum depth of the tree once the kernel
        is reattached.
    stop_probability : ScalarFloat, optional
        The probability of stopping at each level, by default 0.5.
    go_left_probability : ScalarFloat, optional
        The probability of choosing the left child at each
        level, by default 0.5.

    Returns
    -------
    ScalarInt
        The index of the final chosen node (in level order).
    Int[jnp.ndarray, " D"]
        An array of indices representing the path taken, all
        parent indices from the start_idx to the final chosen index
        (in level order), excluding the final index.
    ScalarFloat
        The log probability of reaching the chosen idx. The probability
        is calculated by multiplying the probabilities of stopping at
        at each index along the path and the probabilities
        of choosing the left or right child at each level.
    """

    index_array = jnp.full(max_depth + 1, -1)

    chosen_idx, index_path, num_indices, idx_log_p = _generate_random_path(
        key,
        start_idx,
        max_depth - node_height,
        index_array,
        stop_probability,
        go_left_probability,
    )

    index_path = index_path

    return chosen_idx, index_path, idx_log_p


@jit
def _generate_random_path(
    key: PRNGKeyArray,
    start_idx: ScalarInt,
    max_depth: ScalarInt,
    index_array: Int[jnp.ndarray, " D"],
    stop_probability: ScalarFloat = 0.5,
    go_left_probability: ScalarFloat = 0.5,
) -> tuple[
    ScalarInt,
    Int[jnp.ndarray, " D"],
    ScalarInt,
    ScalarFloat,
]:
    """
    Generate a random path in a binary tree starting from a given index.

    NOTE: This function should be called via the generate_random_path function.

    Parameters
    ----------
    key : jr.PRNGKey
        The random key.
    start_idx : ScalarInt
        The index of the starting node.
    max_depth : ScalarInt
        The maximum depth of the tree.
    index_array :  Int[jnp.ndarray, " D"]
        An empty array to store the path taken, should
        be long enough to store the path to the deepest
        possible node.
    stop_probability : ScalarFloat, optional
        The probability of stopping at each level, by default 0.5.
    go_left_probability : ScalarFloat, optional
        The probability of choosing the left child at each
        level, by default 0.5.

    Returns
    -------
    ScalarInt
        The index of the final chosen node (in level order).
    Int[jnp.ndarray, " D"]
        An array of indices representing the path taken, all
        parent indices from the start_idx to the final chosen index
        (in level order), excluding the final index.
    ScalarInt
        The number of indices in the path (from start_idx to the final
        chosen index).
    ScalarFloat
        The log probability of reaching the chosen idx. The probability
        is calculated by multiplying the probabilities of stopping at
        at each index along the path and the probabilities
        of choosing the left or right child at each level.

    """
    log_stop_probability = jnp.log(stop_probability)
    log_continue_probability = jnp.log(1 - stop_probability)
    log_go_left_probability = jnp.log(go_left_probability)
    log_go_right_probability = jnp.log(1 - go_left_probability)

    def generate_path(state: tuple) -> tuple:
        """Generate a random path in a binary tree."""
        key, idx, index_array, pointer, log_probability, _ = state

        # set the current index in the index array
        index_array = index_array.at[pointer].set(idx)

        key, done_key, direction_key = jr.split(key, 3)

        # evaluate if traversal is done, taking
        # into account if maximum depth is reached
        is_at_max_depth = get_depth(idx) == max_depth
        is_done = jnp.where(
            is_at_max_depth,
            jnp.array(True),
            jr.bernoulli(done_key, p=stop_probability),
        )

        # calculate the probability of stopping at this level,
        # again taking into account if maximum depth is reached
        log_probability = jnp.where(
            is_at_max_depth,
            log_probability,
            jnp.where(
                is_done,
                log_probability + log_stop_probability,
                log_probability + (log_continue_probability),
            ),
        )
        # if the traversal is not done, move to pointer to next node
        pointer = jnp.where(is_done, pointer, pointer + 1)

        # if the traversal is done, return the current index,
        # if not, choose the next index to move to based on the
        # probability of going left or right
        next_idx = jnp.asarray(
            jnp.where(
                is_done,
                idx,
                jnp.where(
                    jr.bernoulli(direction_key, p=go_left_probability),
                    get_child_idx(idx, "l"),
                    get_child_idx(idx, "r"),
                ),
            )
        )
        # check if left or right child was chosen, 0 if right (even idx),
        # 1 if left (odd idx)
        going_left = jnp.mod(next_idx, 2)

        # if traversal is not done, update the probability
        # by multiplying the probability of choosing the left
        # or right child
        log_probability = jnp.where(
            is_done,
            log_probability,
            jnp.where(
                going_left,
                log_probability + log_go_left_probability,
                log_probability + log_go_right_probability,
            ),
        )

        return (key, next_idx, index_array, pointer, log_probability, is_done)

    def condition(state: tuple) -> ScalarBool:
        """Check if generation is done."""
        return ~state[-1]

    initial_state = (key, start_idx, index_array, 0, 0.0, False)

    _, final_idx, index_array, num_indices, final_log_probability, _ = lax.while_loop(
        condition,
        generate_path,
        initial_state,
    )
    return final_idx, index_array, num_indices, final_log_probability


def reconstruct_path(
    start_idx: ScalarInt,
    end_idx: ScalarInt,
    max_depth: ScalarInt,
    stop_probability: ScalarFloat = 0.5,
    go_left_probability: ScalarFloat = 0.5,
) -> tuple[
    Int[jnp.ndarray, " D"],
    ScalarFloat,
]:
    """
    Given the start index and the end index of a path in a binary tree,
    reconstruct the path taken from the start index to the end index and
    calculate the log probability of reaching the end index.
    (This is the inverse of the generate_random_path function.)

    Parameters
    ----------
    start_idx : ScalarInt
        The index of the starting node.
    end_idx : ScalarInt
        The index of the final node.
    max_depth : ScalarInt
        The maximum depth of the tree.
    stop_probability : ScalarFloat, optional
        The probability of stopping at each level, by default 0.5.
    go_left_probability : ScalarFloat, optional
        The probability of choosing the left child at each
        level, by default 0.5.

    Returns
    -------
     Int[jnp.ndarray, " D"]
        An array of indices representing the path taken, all
        parent indices from the start_idx to the final chosen index
        (in level order), excluding the final index.
    ScalarFloat
        The log probability of reaching the chosen idx. The probability
        is calculated by multiplying the probabilities of stopping at
        at each index along the path and the probabilities
        of choosing the left or right child at each level.
    """

    index_array = jnp.full(max_depth + 1, -1)

    path_array, num_indices, log_probability = _reconstruct_path(
        start_idx,
        end_idx,
        max_depth,
        index_array,
        stop_probability,
        go_left_probability,
    )

    # reverse the path array to get the path from start to end
    # (NOTE: To keep array static but only reverse filled part,
    # we reverse the array, then roll last part back to the front).
    path_array = jnp.roll(path_array[::-1], num_indices)
    return path_array, log_probability


@jit
def _reconstruct_path(
    start_idx: ScalarInt,
    end_idx: ScalarInt,
    max_depth: ScalarInt,
    index_array: Int[jnp.ndarray, " D"],
    stop_probability: ScalarFloat = 0.5,
    go_left_probability: ScalarFloat = 0.5,
) -> tuple[
    Int[jnp.ndarray, " D"],
    ScalarInt,
    ScalarFloat,
]:
    """
    Given the start index and the end index of a path in a binary tree,
    reconstruct the path taken from the start index to the end index and
    calculate the log probability of reaching the end index.

    NOTE: This function should be called via the reconstruct_path function.

    Parameters
    ----------
    start_idx : ScalarInt
        The index of the starting node.
    end_idx : ScalarInt
        The index of the final node.
    max_depth : ScalarInt
        The maximum depth of the tree.
    index_array :  Int[jnp.ndarray, " D"]
        An empty array to store the path taken, should
        be long enough to store the path to the deepest
        possible node.
    stop_probability : ScalarFloat, optional
        The probability of stopping at each level, by default 0.5.
    go_left_probability : ScalarFloat, optional
        The probability of choosing the left child at each
        level, by default 0.5.

    Returns
    -------
     Int[jnp.ndarray, " D"]
        An array of indices representing the path taken, all
        parent indices from the start_idx to the final chosen index
        (in level order), excluding the final index.
    ScalarInt
        The number of indices in the path (from start_idx to the final
        chosen index).
    ScalarFloat
        The log probability of reaching the chosen idx. The probability
        is calculated by multiplying the probabilities of stopping at
        at each index along the path and the probabilities
        of choosing the left or right child at each level.

    """
    log_stop_probability = jnp.log(stop_probability)
    log_continue_probability = jnp.log(1 - stop_probability)
    log_go_left_probability = jnp.log(go_left_probability)
    log_go_right_probability = jnp.log(1 - go_left_probability)

    def reconstruct_path(state: tuple) -> tuple:
        """Reconstruct the path from the end index to the start index."""
        current_idx, index_array, pointer, log_probability = state

        # if current index is the end index, don't update the index array
        # and pointer since generate_random_path does not include the end index
        index_array_updated = jnp.where(
            current_idx == end_idx,
            index_array,
            index_array.at[pointer].set(current_idx),
        )
        pointer_updated = jnp.where(
            current_idx == end_idx,
            pointer,
            pointer + 1,
        )

        # check if the current index is at the maximum depth
        # or start, and if its left or right child
        is_at_max_depth = get_depth(current_idx) == max_depth
        is_at_start = current_idx == start_idx
        is_left_node = current_idx % 2  # 0 if right, 1 if left

        # update log probability with the probability of stopping
        # at the current index, taking into account if the current
        # index is at the maximum depth
        log_probability = jnp.where(
            is_at_max_depth,
            log_probability,
            jnp.where(
                current_idx == end_idx,
                log_probability + log_stop_probability,
                log_probability + log_continue_probability,
            ),
        )

        # update log probability with the probability of choosing
        # the left child at the current index, taking into account
        # if the current index is at the start of the path
        log_probability = jnp.where(
            is_at_start,
            log_probability,
            jnp.where(
                is_left_node,
                log_probability + log_go_left_probability,
                log_probability + log_go_right_probability,
            ),
        )

        parent_idx = get_parent_idx(current_idx)

        return (
            parent_idx,
            index_array_updated,
            pointer_updated,
            log_probability,
        )

    def condition(state: tuple) -> ScalarBool:
        """Check if the start index has been reached."""
        return state[0] >= start_idx

    initial_state = (end_idx, index_array, 0, 0.0)

    _, path_array, num_indices, final_log_probability = lax.while_loop(
        condition,
        reconstruct_path,
        initial_state,
    )

    return path_array, num_indices, final_log_probability



import warnings

import beartype.typing as tp
import jax.numpy as jnp
import tensorflow_probability.substrates.jax.bijectors as tfb
from flax.core import FrozenDict

from gallifrey.kernels.atoms import (
    AbstractAtom,
    AbstractOperator,
    LinearAtom,
    PeriodicAtom,
    ProductOperator,
    RBFAtom,
    SumOperator,
)


class KernelLibrary:
    """
    A class to store the kernel library, which contains information
    about what atoms and operators are available for constructing
    the tree kernel.

    Attributes
    ----------
    atoms : tp.Optional[list[AbstractAtom]]
        A list of atoms that are used to construct the tree kernel.
        If None, the default atoms are used. Atoms should inherit
        from the AbstractAtom class in gallifrey.kernels.atoms.
    operators : tp.Optional[list[AbstractOperator]]
        A list of operators that are used to combine the atoms in the
        tree kernel. If None, the default operators are used. Operators
        should inherit from the AbstractOperator class in
        gallifrey.kernels.atoms.
    num_atoms : int
        The number of atoms in the library.
    num_operators : int
        The number of operators in the library.
    library : list[tp.Union[AbstractAtom, AbstractOperator]]
        A list combining the atoms and operators in the library.
    is_operator : jnp.ndarray
        A boolean array to indicate if an entry in the library is an operator
        or not.
    max_atom_parameters : int
        The maximum number of parameters any atom or operator in the library
        takes as input.
    prior_transforms : tp.Optional[dict[str, tfb.Bijector]]
        A dictionary that maps the parameter tags to the corresponding
        transform functions. Transformation functions must transform
        parameter from a standard normal distribution to the desired
        (prior) distribution, and should be implemented via
        tensorflow_probability bijectors. The default transformations
        are used if None. The default transformations are:
        - "positive": Log-normal transformation(mu=0.0, sigma=1.0)
        - "real": Log-normal transformation(mu=0.0, sigma=1.0)
        - "sigmoid": Logit-normal transform(scale=1.0, mu=0.0, sigma=1.0)
        NOTE: To make functions that use this dict jittable, the prior_transforms
        is converted to a FrozenDict, to make it hashable.
        NOTE: The "none" key is reserved for internal use and should not be
        used in the prior_transforms dictionary.
    support_tag_mapping : dict[str, int]
        A dictionary that maps the support tags to integer values.
    support_mapping_array : jnp.ndarray
        An array that contains the support tags as integers for each parameter
        in the library.
    support_transforms : FrozenDict[str, Transformation]
        A (frozen) dict that maps the parameter tags to the corresponding transform
        functions. Transformation functions must transform parameter from a constrained
        space to an unconstrained space. This is used for sampling and optimization.
        NOTE: This should ideally never be touched, unless you know what you are doing.

    """

    def __init__(
        self,
        atoms: tp.Optional[list[AbstractAtom]] = None,
        operators: tp.Optional[list[AbstractOperator]] = None,
        prior_transforms: tp.Optional[dict[str, tfb.Bijector]] = None,
    ):
        """
        Initialize the KernelLibrary. If no atoms or operators are provided,
        a default library is used.

        Parameters
        ----------
        atoms : tp.Optional[list[AbstractAtom]], optional
            A list of atoms that are used to construct the tree kernel.
            If None, the default atoms are used. Atoms should inherit
            from the AbstractAtom class in gallifrey.kernels.atoms, by default None.
        operators : tp.Optional[list[AbstractOperator]], optional
            A list of operators that are used to combine the atoms in the
            tree kernel. If None, the default operators are used. Operators
            should inherit from the AbstractOperator class in
            gallifrey.kernels.atoms, by default None.
        prior_transforms : tp.Optional[dict[str, Transformation]], optional
            A dictionary that maps the parameter tags to the corresponding transform
            functions. Transformation functions must transform parameter from a standard
            normal distribution to the desired (prior) distribution, and should be
            implemented via tensorflow_probability bijectors.
            The default transformations are used if None, by default None. The default
            transformations are:
            - "positive": Log-normal transformation(mu=0.0, sigma=1.0)
            - "real": Log-normal transformation(mu=0.0, sigma=1.0)
            - "sigmoid": Logit-normal transform(scale=1.0, mu=0.0, sigma=1.0)
            NOTE: To make 'transformation' function (see gallifrey.kernels.prior)
            jitable, the prior_transforms must be hashable. This is why we use
            FrozenDict instead of dict.

        """
        self.atoms: list[AbstractAtom] = (
            atoms
            if atoms is not None
            else [
                LinearAtom(),
                PeriodicAtom(),
                RBFAtom(),
            ]
        )

        self.operators: list[AbstractOperator] = (
            operators
            if operators is not None
            else [
                SumOperator(),
                ProductOperator(),
            ]
        )

        self.num_atoms = len(self.atoms)
        self.num_operators = len(self.operators)

        # construct the library by combining the atoms and operators
        self.library = self.atoms + self.operators

        # create boolean array to indicate if entry is operator or not
        self.is_operator = jnp.array(
            [False] * len(self.atoms) + [True] * len(self.operators)
        )

        # get the maximum number of parameters any atom or operator takes
        self.max_atom_parameters = max([item.num_parameter for item in self.atoms])

        # transformation functions from normal distribution to prior distribution
        # NOTE: names are inherited from GPJax, but we don't use the same
        # transformations (which is why sigmoid is the name for the logit-normal)
        if prior_transforms is None:
            self.prior_transforms: FrozenDict[str, tfb.Bijector] = FrozenDict(
                {
                    # this is the transformation y = exp(mu + sigma * z),
                    # with mu = 0 and sigma = 1,
                    # if z ~ normal(0, 1) then y ~ log-normal(mu, sigma)
                    "real": tfb.Chain(
                        [
                            tfb.Exp(),
                            tfb.Shift(jnp.array(0.0)),
                            tfb.Scale(jnp.array(1.0)),
                        ]
                    ),
                    "positive": tfb.Chain(
                        [
                            tfb.Exp(),
                            tfb.Shift(jnp.array(0.0)),
                            tfb.Scale(jnp.array(1.0)),
                        ]
                    ),
                    # this is the transformation y = 1/(1 + exp(-(mu + sigma * z))),
                    # with mu = 0 and sigma = 1,
                    # if z ~ normal(0, 1) then y ~ logit-normal(mu, sigma)
                    "sigmoid": tfb.Chain(
                        [
                            tfb.Sigmoid(
                                low=jnp.array(0.0),
                                high=jnp.array(0.95),  # to avoid numerical issues
                            ),
                            tfb.Shift(jnp.array(0.0)),
                            tfb.Scale(jnp.array(0.0)),
                        ]
                    ),
                    # identity transformation for parameter that do not fall under
                    # the above categories
                    "none": tfb.Identity(),
                }
            )

        else:
            if "none" in prior_transforms.keys():
                raise ValueError(
                    "'prior_transforms' should not contain the key 'none', "
                    "as it is reserved for internal use."
                )
            prior_transforms["none"] = tfb.Identity()
            self.prior_transforms = FrozenDict(prior_transforms)

        self.support_mapping_array, self.support_tag_mapping = (
            self.get_support_mapping()
        )

        # besides the prior transforms, we also perform transforms
        # between a constrained and unconstrained space for fitting/sampling,
        # this probably should never be touched unless you implement a new
        # kernel with new constraints
        self.support_transforms: FrozenDict[str, tfb.Bijector] = FrozenDict(
            {
                "positive": tfb.Softplus(),
                "real": tfb.Identity(),
                "sigmoid": tfb.Sigmoid(low=0.0, high=0.95),
                "lower_triangular": tfb.FillTriangular(),
                "none": tfb.Identity(),
            }
        )

        self._check_tags()

    def __repr__(self) -> str:
        """
        Get the (technical) string representation of the KernelLibrary.

        Returns
        -------
        str
            The (technical) string representation of the KernelLibrary.
        """

        return (
            f"KernelLibrary(\n  Atoms={self.atoms},\n   "
            f"operators={self.operators}\n)"
        )

    def __str__(self) -> str:
        """
        Get the (simplified) string representation of the KernelLibrary.

        Returns
        -------
        str
            The (simplified) string representation of the KernelLibrary.
        """
        try:
            atom_names: list[tp.Any] = [atom.name for atom in self.atoms]
        except Exception:
            atom_names = self.atoms
        try:
            operator_names: list[tp.Any] = [
                operator.name for operator in self.operators
            ]
        except Exception:
            operator_names = self.operators

        return (
            f"KernelLibrary(\n  Atoms={atom_names},\n   "
            f"operators={operator_names}\n)"
        )

    def __len__(self) -> int:
        """
        Get the number of items in the library.

        Returns
        -------
        int
            The number of items in the library.
        """
        return len(self.library)

    def get_support_mapping(self) -> tuple[jnp.ndarray, dict[str, int]]:
        """
        Create a mapping between support tags and the integer values, then
        create an array that contains the support tags for each parameter
        in the library.
        (Used for jittable sampling and parameter transformations)

        Returns
        -------
        support_mapping_array : jnp.ndarray
            An array that contains the support tags for each parameter
            in the library, with shape (len(self), self.max_atom_parameters),
            enocded as integers.
        support_tag_mapping : dict[str, int]
            A dictionary that maps the support tags to the corresponding
            integer values.
        """

        # create dict that maps string tags to integer values
        support_tag_mapping = {
            support_tag: i for i, support_tag in enumerate(self.prior_transforms.keys())
        }
        support_tag_mapping["none"] = -1

        # create array that contains the support tag integers for each parameter
        # in the library
        support_mapping_array = jnp.full(
            (len(self), self.max_atom_parameters), support_tag_mapping["none"]
        )

        for i in range(len(self)):
            for j in range(self.library[i].num_parameter):
                support_mapping_array = support_mapping_array.at[i, j].set(
                    support_tag_mapping[self.library[i].parameter_support[j]]
                )
        return support_mapping_array, support_tag_mapping

    def _check_tags(self) -> None:
        """
        Check if the parameters of the atoms have constraints that are not yet
        implemented in the library. If so, raise a warning.

        """
        parameter_support_tags = {
            atom.name: atom.parameter_support for atom in self.atoms
        }

        for atom_name, tags in parameter_support_tags.items():
            if not all(tag in self.prior_transforms.keys() for tag in tags):
                warnings.warn(
                    f"Atom {atom_name!r} has parameter tags that are not "
                    "in the prior_transforms dictionary. Tag will be ignored. "
                    "This will lead to problems when performing parameter "
                    "transformations. Add the missing tags to the prior_transforms"
                    "by manually passing the prior_transforms dictionary to the "
                    "KernelLibrary constructor."
                )

        if not all(tag in self.support_transforms.keys() for tag in tags):
            warnings.warn(
                f"Kernel {atom_name!r} has parameter tags that are not "
                "in the support_transforms dictionary. This will lead to problems "
                "in the sampling/optmization if used for a KernelTree object. "
                "Add the missing tags to the by manually overriding the "
                "support_transforms dictionary to the KernelLibrary constructor.\n"
                "THIS IS VERY EXPERIMENTAL AND NOT RECOMMENDED."
            )



import jax.numpy as jnp
from jaxtyping import Float


class Dataset:
    """
    A class to hold the data.

    Attributes
    ----------
    x : Float[jnp.ndarray, " D"]
        The input x data (1D).
    y : Float[jnp.ndarray, " D"]
        The input y data (1D).
    n : int
        The number of data points.

    """

    def __init__(
        self,
        x: Float[jnp.ndarray, " D"],
        y: Float[jnp.ndarray, " D"],
    ):
        """
        Initialize the Dataset instance.

        Parameters
        ----------
        x : Float[jnp.ndarray, " D"]
            The input x data.
        y : Float[jnp.ndarray, " D"]
            The input y data.
        """
        self.x = jnp.asarray(x)
        self.y = jnp.asarray(y)

        self._validate_input()

    @property
    def n(self) -> int:
        """
        Get the number of data points.

        Returns
        -------
        int
            The number of data points.
        """
        return self.x.shape[0]

    def _validate_input(self) -> None:
        """
        Check that the number of x and y values are the same.

        Raises
        ------
        ValueError
            If the number of x and y values are not the same.
        """

        if self.x.shape[0] != self.y.shape[0]:
            raise ValueError("The length of x and y must be the same.")

    def __repr__(self) -> str:
        """
        Return the string representation of the Dataset instance.

        Returns
        -------
        str
            The representation of the Dataset instance.
        """
        return f"Dataset(n={self.n},\n x={self.x},\n y={self.y})"

    def __str__(self) -> str:
        """
        Simplified string representation of the Dataset instance.

        Returns
        -------
        str
            The simplified string representation.
        """
        return f"Dataset(n={self.n})"



from abc import ABC, abstractmethod
from typing import Literal

import jax.numpy as jnp
from jax import vmap
from jaxtyping import Float


class AbstractAtom(ABC):
    """
    Abstract base class for defining atom kernels
    (basic kernel functions).

    Atoms are the fundamental building blocks for constructing more complex kernel
    structure. This class defines the interface that all atom kernels must
    implement within the `gallifrey` library.

    Attributes
    ----------
    name : str
        A descriptive name for the atom kernel (e.g., "RBF", "Linear"). This name
        is used for identification and representation purposes.
    num_parameter : int
        The number of trainable parameters associated with this atom kernel.
        For example, an RBF kernel might have a lengthscale and a variance parameter,
        so `num_parameter` would be 2.
    parameter_support : list[Literal["positive", "real", "sigmoid", "none"]]
        A list defining the support constraints for each parameter of the atom kernel.
        Each element in the list corresponds to a parameter and specifies its
        constraint:
        - `"positive"`: Parameter must be positive (e.g., lengthscale, variance).
        - `"real"`: Parameter can be any real number (unconstrained).
        - `"sigmoid"`: Parameter is constrained to the range (0, 1) via a sigmoid
          function.
        - `"none"`: No specific constraint is applied.
        The length of this list should be equal to `num_parameter`.
    parameter_names : list[str]
        A list of names for each parameter of the atom kernel. These names should be
        descriptive and correspond to the parameters in the order defined.
        The length of this list should be equal to `num_parameter`.

    Methods
    -------
    __call__(x, y, params)
        Abstract method to compute the kernel value between two input points `x`
        and `y`. This method must be implemented by concrete atom kernel classes.
    __str__()
        Returns the name of the atom kernel as a string.
    cross_covariance(x, y, params)
        Computes the cross-covariance matrix between two sets of input vectors `x` and
        `y`. Innput must be a 1D vector. The output is a 2D matrix.
        This method uses `jax.vmap` for efficient computation over batched inputs.
    gram(x, params)
        Computes the Gram matrix (covariance matrix) for a vector of input points `x`.
        This is a special case of `cross_covariance` where both input sets are the same.

    """

    name: str
    num_parameter: int
    parameter_support: list[Literal["positive", "real", "sigmoid", "none"]]
    parameter_names: list[str]

    @abstractmethod
    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "..."]:
        pass

    def __str__(self) -> str:
        return self.name

    def cross_covariance(
        self,
        x: Float[jnp.ndarray, " M"] | Float[jnp.ndarray, "M 1"],
        y: Float[jnp.ndarray, " N"] | Float[jnp.ndarray, "N 1"],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "M N"]:
        return vmap(lambda x_: vmap(lambda y_: self(x_, y_, params))(y))(x)

    def gram(
        self,
        x: Float[jnp.ndarray, " D"] | Float[jnp.ndarray, "D 1"],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "D D"]:
        return self.cross_covariance(x, x, params)


class AbstractOperator(ABC):
    """
    An abstract class to define the operators (combination
    functions) in the kernel structure. The operators are
    used to combine the atoms in the kernel computations.

    """

    name: str
    num_parameter: int = 0
    parameter_support: list = []
    parameter_names: list = []

    @abstractmethod
    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
    ) -> Float[jnp.ndarray, "..."]:
        pass


class SumOperator(AbstractOperator):
    """Sum Operator to combine two kernel functions."""

    name = "+"

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
    ) -> Float[jnp.ndarray, "..."]:
        return jnp.asarray(x + y).squeeze()


class ProductOperator(AbstractOperator):
    """Product Operator to combine two kernel functions."""

    name = "*"

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
    ) -> Float[jnp.ndarray, "..."]:
        return jnp.asarray(x * y).squeeze()


class ConstantAtom(AbstractAtom):
    """Constant atom."""

    name = "Constant"
    num_parameter = 1
    parameter_support = ["real"]
    parameter_names = ["constant"]

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "..."]:
        constant = params[0]
        return jnp.asarray(constant).astype(jnp.asarray(x).dtype).squeeze()


class RBFAtom(AbstractAtom):
    """Radial Basis Function/Squared Exponential atom."""

    name = "RBF"
    num_parameter = 2
    parameter_support = ["positive", "positive"]
    parameter_names = ["lengthscale", "variance"]

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "..."]:
        lengthscale = params[0]
        variance = params[1]
        d_squared = jnp.square(x - y) / jnp.square(lengthscale)
        k = variance * jnp.exp(-0.5 * d_squared)
        return k.squeeze()


class Matern12Atom(AbstractAtom):
    """Matern 1/2 atom."""

    name = "Matern12"
    num_parameter = 2
    parameter_support = ["positive", "positive"]
    parameter_names = ["lengthscale", "variance"]

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "..."]:
        lengthscale = params[0]
        variance = params[1]
        tau = jnp.abs(x - y) / lengthscale
        k = variance * jnp.exp(-tau)
        return k.squeeze()


class Matern32Atom(AbstractAtom):
    """Matern 3/2 atom."""

    name = "Matern32"
    num_parameter = 2
    parameter_support = ["positive", "positive"]
    parameter_names = ["lengthscale", "variance"]

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "..."]:
        lengthscale = params[0]
        variance = params[1]
        tau = jnp.sqrt(3.0) * jnp.abs(x - y) / lengthscale
        k = variance * (1.0 + tau) * jnp.exp(-tau)
        return k.squeeze()


class Matern52Atom(AbstractAtom):
    """Matern 5/2 atom."""

    name = "Matern52"
    num_parameter = 2
    parameter_support = ["positive", "positive"]
    parameter_names = ["lengthscale", "variance"]

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "..."]:
        lengthscale = params[0]
        variance = params[1]
        tau = jnp.sqrt(5.0) * jnp.abs(x - y) / lengthscale
        k = variance * (1.0 + tau + jnp.square(tau) / 3.0) * jnp.exp(-tau)
        return k.squeeze()


class PeriodicAtom(AbstractAtom):
    """Periodic atom."""

    name = "Periodic"
    num_parameter = 3
    parameter_support = ["positive", "positive", "positive"]
    parameter_names = ["lengthscale", "variance", "period"]

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "..."]:
        lengthscale = params[0]
        variance = params[1]
        period = params[2]
        sine_squared = (jnp.sin(jnp.pi * (x - y) / period) / lengthscale) ** 2
        k = variance * jnp.exp(-0.5 * sine_squared)
        return k.squeeze()


class PoweredExponentialAtom(AbstractAtom):
    """Powered Exponential atom."""

    name = "PoweredExponential"
    num_parameter = 3
    parameter_support = ["positive", "positive", "sigmoid"]
    parameter_names = ["lengthscale", "variance", "power"]

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "..."]:
        lengthscale = params[0]
        variance = params[1]
        power = params[2]
        tau = jnp.abs(x - y) / lengthscale
        k = variance * jnp.exp(-jnp.power(tau, power))
        return k.squeeze()


class RationalQuadraticAtom(AbstractAtom):
    """Rational Quadratic atom."""

    name = "RationalQuadratic"
    num_parameter = 3
    parameter_support = ["positive", "positive", "positive"]
    parameter_names = ["lengthscale", "variance", "alpha"]

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "..."]:
        lengthscale = params[0]
        variance = params[1]
        alpha = params[2]
        tau = jnp.abs(x - y) / lengthscale
        k = variance * (1 + 0.5 * jnp.square(tau) / alpha) ** (-alpha)
        return k.squeeze()


class WhiteAtom(AbstractAtom):
    """White atom."""

    name = "White"
    num_parameter = 1
    parameter_support = ["positive"]
    parameter_names = ["variance"]

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "..."]:
        variance = params[0]
        k = jnp.equal(x, y) * variance
        return k.squeeze()


class LinearAtom(AbstractAtom):
    """(Non-stationary) linear atom."""

    name = "Linear"
    num_parameter = 1
    parameter_support = ["positive"]
    parameter_names = ["variance"]

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "..."]:
        variance = params[0]
        k = variance * (x * y)
        return jnp.asarray(k).squeeze()


class LinearWithShiftAtom(AbstractAtom):
    """(Non-stationary) linear atom."""

    name = "Linear"
    num_parameter = 3
    parameter_support = ["positive", "positive", "real"]
    parameter_names = ["bias", "variance", "shift"]

    def __call__(
        self,
        x: Float[jnp.ndarray, "..."],
        y: Float[jnp.ndarray, "..."],
        params: Float[jnp.ndarray, " P"],
    ) -> Float[jnp.ndarray, "..."]:
        bias = params[0]
        variance = params[1]
        shift = params[2]
        k = bias + variance * ((x - shift) * (y - shift))
        return jnp.asarray(k).squeeze()



from __future__ import annotations

from functools import partial

import beartype.typing as tp
import jax.numpy as jnp
from equinox.internal import while_loop as bounded_while_loop
from flax import nnx
from jax import jit, lax
from jaxtyping import Bool, Float, Int

from gallifrey.kernels.atoms import AbstractAtom, AbstractOperator
from gallifrey.kernels.library import KernelLibrary
from gallifrey.parameter import KernelParameter
from gallifrey.utils.tree_helper import (
    calculate_max_leaves,
    calculate_max_nodes,
    calculate_max_stack_size,
    get_child_idx,
    get_depth,
    get_parameter_leaf_idx,
)
from gallifrey.utils.typing import ScalarBool, ScalarFloat, ScalarInt


class TreeKernel(nnx.Module):
    """
    A kernel class to evaluate a tree-like kernel expression. The tree expression
    is a jnp array, with the tree structure encoded as a level-order array.
    Similar to NestedCombinationKernel in GPJax, but with explicit tree
    structure.

    Assume for example that the tree expression is [2, 1, 2, -1, -1, 0, 1]
    and the kernel library is [RBFAtom, LinearAtom, SumAtom, ProductAtom]. In this case
    -1 -> empty node,
    0 -> RBFKernel,
    1 -> LinearKernel,
    2 -> Sum,
    3 -> Product.
    The is_operator array should be [False, False, True, True].
    The tree expression would then be evaluated to
    Sum(Linear(x, y), Sum(RBF(x,y), Linear(x,y)).

    The same kernel could be constructed using NestedCombinationKernel, but the
    tree structure would not be explicit.

    Attributes
    (NOTE: A lot of values are initialized as nnx.Variable, use .value to access the
    actual value.)
    ----------
    tree_expression : Int[jnp.ndarray, " D"]
        The tree expression, a level-order array describing the tree structure.
        Negative values indicate empty nodes.
    max_nodes : ScalarInt
        The maximum number of nodes in the tree expression.
    max_depth : int
        The maximum depth of the tree expression.
    max_stack : ScalarInt
        The maximum size of the stack used to evaluate the tree expression,
        which is equal to the maximum depth of the tree expression + 1.
    max_leaves : ScalarInt
        The maximum number of leaves in the tree expression.
    max_atom_parameters : ScalarInt
        The maximum number of parameters of any atom in the atom library. Comes
        from the kernel library.
    max_total_parameters : ScalarInt
        The maximum number of parameters in the tree expression (max_leaves
        * max_atom_parameters).
    num_datapoints : ScalarInt
        The number of datapoints in the input training data. (This is needed to
        correctly construct the gram matrix, in a jit-compatible way.)
    root_idx : ScalarInt
        The index of the root node in the tree expression.
    atoms : tuple[AbstractAtom]
        The atoms in the kernel library.
    operators : tuple[AbstractOperator]
        The operators in the kernel library.
    num_atoms : ScalarInt
        The number of atoms in the kernel library.
    num_operators : ScalarInt
        The number of operators in the kernel library.
    is_operator : Bool[jnp.ndarray, " D"]
        A boolean array indicating whether the node in the tree expression is an
        operator or not. Comes from the kernel library.
    post_order_expression : Int[jnp.ndarray, " D"]
        The tree_expression in post-order traversal notation. Used for efficient
        evaluation of the tree expression. Negative values indicate empty nodes.
    post_level_map : Int[jnp.ndarray, " D"]
        A map from the post-order index to the level-order index. The value at
        a given index in the post-order expression corresponds to the index of
        the same node in the level-order expression. Negative values indicate
        empty nodes.
    num_nodes : ScalarInt
        The actual number of nodes in the tree expression.
    node_sizes : Int[jnp.ndarray, " D"]
        An array containing the sizes of the nodes in the tree. The size of a node
        is the number of nodes in the subtree rooted at that node. The size of a
        given node is the value at its index in the level-order expression.
    node_heights : Int[jnp.ndarray, " D"]
        An array containing the heights of the nodes in the tree. The height of a node
        is the length of the longest path from the node to a leaf node (i.e. the heigh
        of a leaf node is 0). The height of a given node is the value at its index in
        the level-order expression.
    parameters : KernelParameter(Float[jnp.ndarray, "M N"])
        KernelParameter instance, that holds a 2D array of kernel parameters with shape
        (max_leaves, max_atom_parameters). The parameters are used to evaluate the tree
        kernel.
    leaf_level_map : Int[jnp.ndarray, " N"]
        A map from the leaf index to the level-order index. The leaf index i corresponds
        to the i-th entry in the parameters 0th axis. The value of the leaf_level_map
        at index i is the index of the node with these parameters in the level-order
        expression.
        NOTE: The leaf index is decided based on the depth of the tree, and might not be
        in a simple order. See 'gallifrey.utils.tree_helper.get_parameter_leaf_idx' for
        more information.

    """

    def __init__(
        self,
        tree_expression: Int[jnp.ndarray, " D"],
        kernel_library: KernelLibrary,
        max_depth: int,
        num_datapoints: int,
        root_idx: Int[jnp.ndarray, ""] = jnp.array(0),
    ):
        """
        Initialize a tree kernel.

        Parameters
        ----------
        tree_expression : Int[jnp.ndarray, " D"]
            The tree expression, a level-order expression of the tree.
        kernel_library : KernelLibrary
            An instance of the KernelLibrary class, containing the atoms
            and operators to evaluate the tree expression.
        max_depth : int
            The maximum depth of the tree expression. The length of the
            tree expression should be 2^(max_depth+1) - 1.
        num_datapoints : int, optional
            The number of datapoints in the input data. (This is needed to
            correctly construct the gram matrix.)
        root_idx : Int[jnp.ndarray, ""], optional
            The index of the root node in the tree expression. By default 0.

        """
        self.init_tree(
            tree_expression,
            kernel_library,
            max_depth,
            num_datapoints,
            root_idx,
        )
        self.init_parameters()

    def init_tree(
        self,
        tree_expression: Int[jnp.ndarray, " D"],
        kernel_library: KernelLibrary,
        max_depth: int,
        num_datapoints: int,
        root_idx: Int[jnp.ndarray, ""] = jnp.array(0),
    ) -> None:
        """
        Initialize the base attributes of the tree kernel, which means
        setting the
        - tree_expression,
        - max_nodes,
        - root_idx,
        - max_depth,
        - is_operator (from the kernel library),
        - atom_library (from the kernel library),
        - max_atom_parameters (from the kernel library),
        - post_order_expression,
        - post_level_map,
        - num_nodes,
        - node_sizes, and
        - node_heights.

        The pre-order traversal expression and map attributes are initialized
        but set to None. They get filled when calling the 'display' method.

        Parameters
        ----------
        tree_expression : Int[jnp.ndarray, " D"]
            The tree expression, a level-order expression of the tree.
        kernel_library : tp.Optional[KernelLibrary], optional
            An instance of the KernelLibrary class, containing the atoms and
            operators to evaluate the tree expression.
        max_depth : int
            The maximum depth of the tree expression. The length of the
            tree expression should be 2^(max_depth+1) - 1.
        num_datapoints : int, optional
            The number of datapoints in the input data. (This is needed to
            correctly construct the gram matrix.)
        root_idx : Int[jnp.ndarray, ""], optional
            The index of the root node in the tree expression. By default 0.

        """
        # set base attributes
        self.tree_expression = nnx.Variable(tree_expression)

        self.max_depth = nnx.Variable(max_depth)
        self.max_stack = nnx.Variable(calculate_max_stack_size(max_depth))
        self.max_nodes = nnx.Variable(calculate_max_nodes(max_depth))
        self.max_leaves = nnx.Variable(calculate_max_leaves(max_depth))

        self.num_datapoints = nnx.Variable(num_datapoints)

        self.root_idx = nnx.Variable(root_idx)

        # check if the root index is valid
        # if self.root_idx.value > self.max_nodes - 1:
        #     raise ValueError(
        #         "Root index must be less than the maximum number of nodes."
        #     )

        # get the kernel library attributes
        self.atoms = tuple(kernel_library.atoms)
        self.operators = tuple(kernel_library.operators)
        self.num_atoms = nnx.Variable(kernel_library.num_atoms)
        self.num_operators = nnx.Variable(kernel_library.num_operators)
        self.is_operator = nnx.Variable(kernel_library.is_operator)
        self.max_atom_parameters = nnx.Variable(kernel_library.max_atom_parameters)
        self.max_total_parameters = nnx.Variable(
            self.max_leaves.value * self.max_atom_parameters.value  # type: ignore
        )

        # run the post-order traversal and get node metrics
        (
            post_order_expression,
            num_nodes,
            post_level_map,
            node_sizes,
            node_heights,
            leaf_level_map,
        ) = self.get_post_order_and_node_metrics()

        self.post_order_expression = nnx.Variable(post_order_expression)
        self.post_level_map = nnx.Variable(post_level_map)
        self.num_nodes = nnx.Variable(num_nodes)
        self.node_sizes = nnx.Variable(node_sizes)
        self.node_heights = nnx.Variable(node_heights)
        self.leaf_level_map = nnx.Variable(leaf_level_map)

    def init_parameters(self) -> None:
        """
        Initialize the parameters of the tree kernel. The parameters are
        a KernelParameter instance, which holds a 2D array of kernel parameters
        with shape (max_leaves, max_atom_parameters).
        """
        # initialise parameters
        self.parameters = KernelParameter(
            jnp.ones(
                (
                    self.max_leaves.value,
                    self.max_atom_parameters.value,
                )
            )
        )

    def __call__(
        self,
        x: Float[jnp.ndarray, " D"],
        y: Float[jnp.ndarray, " D"],
    ) -> ScalarFloat:
        """
        Evaluate the tree kernel by traversing the tree expression.

        Parameters
        ----------
        x : Float[jnp.ndarray, " D"]
            Input x data.
        y : Float[jnp.ndarray]
            Input y data.

        Returns
        -------
        ScalarFloat
            The kernel value.
        """
        return _evaluate_tree(
            jnp.atleast_1d(x),
            jnp.atleast_1d(y),
            self.post_order_expression.value,
            self.post_level_map.value,
            self.is_operator.value,
            self.parameters.value,
            self.atoms,
            self.operators,
            (1, 1),
            int(self.num_atoms.value),
            int(self.max_depth.value),
        ).squeeze()

    def _tree_viz(self, include_parameters: bool = True) -> str:
        """
        Helper method for __str__ with optional parameter to include the
        kernel parameters in the visualization.

        Parameters
        ----------
        include_parameters : bool, optional
            Whether to include the kernel parameters in the visualization.

        Returns
        -------
        str
            A string containing the visual representation of the tree kernel.
        """

        pre_order_expression, num_nodes, pre_level_map = self.get_pre_order()

        description = tree_visualization(
            pre_order_expression[:num_nodes],
            pre_level_map,
            self.atoms + self.operators,
            self.parameters.value,
            self.is_operator.value,
            self.root_idx.value,
            self.max_depth.value,
            include_parameters=include_parameters,
            print_str=False,
            return_str=True,
        )
        assert isinstance(description, str)
        return description

    def __str__(self) -> str:
        """
        Construct a visual representation of the tree kernel.
        See tree_visualization for more information.

        Returns
        -------
        str
            A string containing the visual representation of the tree kernel (if
            return_str is True).
        """
        return self._tree_viz(include_parameters=True)

    def cross_covariance(
        self,
        x: Float[jnp.ndarray, " M"] | Float[jnp.ndarray, " M 1"],
        y: Float[jnp.ndarray, " N"] | Float[jnp.ndarray, " N 1"],
    ) -> Float[jnp.ndarray, "M N"]:
        """
        Calculate the cross-covariance matrix between x and y
        for the tree kernel.

        Parameters
        ----------
        x : Float[jnp.ndarray, " M"] | Float[jnp.ndarray, " M 1"]
            Input x data.
        y : Float[jnp.ndarray, " N"] | Float[jnp.ndarray, " N 1"]
            Input y data.


        Returns
        -------
        Float[jnp.ndarray, "M N"]
            The cross-covariance matrix.
        """
        return _evaluate_tree(
            x,
            y,
            self.post_order_expression.value,
            self.post_level_map.value,
            self.is_operator.value,
            self.parameters.value,
            tuple([atom.cross_covariance for atom in self.atoms]),
            self.operators,
            (int(len(x)), int(len(y))),
            int(self.num_atoms.value),
            int(self.max_depth.value),
        )

    def gram(
        self,
        x: Float[jnp.ndarray, " M"] | Float[jnp.ndarray, " M 1"],
    ) -> Float[jnp.ndarray, "M M"]:
        """
        Calculate the gram matrix for a given input x using
        the tree kernel.

        Parameters
        ----------
        x : Float[jnp.ndarray, " D"] | Float[jnp.ndarray, " D 1"]
            Input x data.

        Returns
        -------
        Float[jnp.ndarray, "M M"]
            The gram matrix.
        """
        return self.cross_covariance(x, x)

    def _gram_train(
        self,
        x: Float[jnp.ndarray, " M"] | Float[jnp.ndarray, " M 1"],
    ) -> Float[jnp.ndarray, "M M"]:
        """
        Calculate the gram matrix for a given input x using
        the tree kernel. This version has the number of datapoints
        fixed to the initial number of datapoints in the training data,
        and is used for evaluation in a jit-compatible way.

        Returns
        -------
        Float[jnp.ndarray, "M M"]
            The gram matrix.
        """

        return _evaluate_tree(
            x,
            x,
            self.post_order_expression.value,
            self.post_level_map.value,
            self.is_operator.value,
            self.parameters.value,
            tuple([atom.cross_covariance for atom in self.atoms]),
            self.operators,
            (int(self.num_datapoints.value), int(self.num_datapoints.value)),
            int(self.num_atoms.value),
            int(self.max_depth.value),
        )

    def get_pre_order(
        self,
    ) -> tuple[Int[jnp.ndarray, " D"], ScalarInt, Int[jnp.ndarray, " D"]]:
        """
        Get the pre-order traversal expression of the tree, as well as
        the index of the last node in the tree and a map from the pre-order
        index to the level-order index.


        Returns
        -------
        Int[jnp.ndarray, " D"]
            The pre-order traversal expression of the tree.
        ScalarInt
            The total number of nodes in the tree.
        Int[jnp.ndarray, " D"]
            A map from the pre-order index to the level-order index. The value at a
            given index in the pre-order expression corresponds to the index of the
            same node in the level-order expression
        """

        initial_expression = jnp.full(self.max_nodes.value, -1)
        initial_expression_pointer = 0
        pre_level_map = jnp.copy(initial_expression)
        initial_stack = (
            jnp.copy(initial_expression).at[0].set(self.root_idx)
        )  # fill with root
        initial_stack_pointer = 0

        initial_state = (
            initial_expression,
            initial_expression_pointer,
            pre_level_map,
            initial_stack,
            initial_stack_pointer,
        )

        return level_order_to_pre_order(
            self.tree_expression.value,
            initial_state,
            self.max_nodes.value,
        )

    def get_post_order_and_node_metrics(
        self,
    ) -> tuple[
        Int[jnp.ndarray, " D"],
        ScalarInt,
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
    ]:
        """
        Get the post-order traversal expression of the tree, the output is
        - the post-order traversal expression of the tree,
        - the total number of nodes in the tree,
        - a map from the post-order index to the level-order index,
        - an array containing the sizes of the nodes in the tree (located at the
          level-order index),
        - an array containing the heights of the nodes in the tree (located at
          the level-order index).
        - a map from the leaf index to the level-order index,
        - the total number of leaves in the tree.

        (Unlike the get_pre_order method, this method also returns the
        node sizes, heights, leaf level map, and number of leaves.)

        Returns
        -------
        Int[jnp.ndarray, " D"]
            The post-order traversal expression of the tree.
        ScalarInt
            The total number of nodes in the tree.
        Int[jnp.ndarray, " D"]
            A map from the post-order index to the level-order index. The value at a
            given index in the post-order expression corresponds to the index of the
            same node in the level-order expression
        Int[jnp.ndarray, " D"]
            An array containing the sizes of the nodes in the tree. The size of a node
            is the number of nodes in the subtree rooted at that node. The size of a
            given node is the value at its index in the level-order expression.
        Int[jnp.ndarray, " D"]
            An array containing the heights of the nodes in the tree. The height of a
            node is the length of the longest path from the node to a leaf node
            (i.e. the height of a leaf node is 0). The height of a given
            node is the value at its index in the level-order expression.
        Int[jnp.ndarray, " D"]
            A map from the leaf index to the level-order index. The value at a given
            index in the leaf level map corresponds to the index of the node with
            these parameters in the level-order expression.
            NOTE: The leaf index is decided based on the depth of the tree, and might
            not be in a simple order. See
            'gallifrey.utils.tree_helper.get_parameter_leaf_idx' for more information.
        """
        initial_expression = jnp.full(self.max_nodes.value, -1)
        initial_expression_pointer = 0
        post_level_map = jnp.copy(initial_expression)
        initial_stack = (
            jnp.copy(initial_expression).at[0].set(self.root_idx)
        )  # fill with root
        initial_stack_pointer = 0
        last_processed_idx = -1
        node_sizes = jnp.full(self.max_nodes, 0)
        node_heights = jnp.copy(node_sizes)
        leaf_level_map = jnp.full(self.max_leaves, -1)

        initial_state = (
            initial_expression,
            initial_expression_pointer,
            post_level_map,
            initial_stack,
            initial_stack_pointer,
            last_processed_idx,
            node_sizes,
            node_heights,
            leaf_level_map,
        )

        return level_order_to_post_order_and_metrics(
            self.tree_expression.value,
            initial_state,
            self.is_operator.value,
            self.max_nodes.value,
            self.max_depth.value,
        )

    def print_atoms(self) -> None:
        """
        Print the atoms in the atom library, together with their parameter names.
        """
        for atom in self.atoms:
            print(f"{atom.name}: {atom.parameter_names}")

    def display(self) -> None:
        print(self.__str__())


@jit
def level_order_to_pre_order(
    level_order: Int[jnp.ndarray, " D"],
    initial_state: tuple[
        Int[jnp.ndarray, " D"],
        ScalarInt,
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
        ScalarInt,
    ],
    max_nodes: ScalarInt,
) -> tuple[Int[jnp.ndarray, " D"], ScalarInt, Int[jnp.ndarray, " D"]]:
    """
    Convert a level-order expression of a tree to a pre-order expression.
    Also returns the total number of nodes in the tree. The
    pre-order expression is used to visualize the tree kernel.

    NOTE: This function should be called via the get_pre_order method of the
    TreeKernel.

    Parameters
    ----------
    level_order : Int[jnp.ndarray, " D"]
        The level-order expression of the tree.
    initial_state : tuple
        The initial state, containing the pre-order expression, the
        pre-order pointer, the pre_level_map, the stack, and the stack pointer.
    max_nodes : ScalarInt
        The maximum number of nodes in the tree expression.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The pre-order expression of the tree.
    ScalarInt
        The total number of nodes in the tree.
    Int[jnp.ndarray, " D"]
        A map from the pre-order index to the level-order index. The value at a given
        index in the pre-order expression corresponds to the index of the same node in
        the level-order expression.

    """

    def traverse_tree(state: tuple) -> tuple:
        """Traverse the tree expression in level-order and convert it to
        pre-order."""
        pre_order, pre_order_pointer, pre_level_map, stack, stack_pointer = state

        # pop the top of the stack
        level_order_idx = stack[stack_pointer]
        stack_pointer = stack_pointer - 1

        # add the current node to the pre-order output
        new_pre_order = pre_order.at[pre_order_pointer].set(
            level_order[level_order_idx]
        )
        new_pre_order_pointer = pre_order_pointer + 1
        # update the pre_level_map
        new_pre_level_map = pre_level_map.at[pre_order_pointer].set(level_order_idx)

        # check if the current node is a leaf
        right_child_idx = get_child_idx(level_order_idx, "l")
        is_leaf = jnp.logical_or(
            right_child_idx > max_nodes - 1,
            level_order[right_child_idx] == -1,
        )

        def push(stack_and_pointer: tuple) -> tuple:
            """If the node is not a leaf, push the new
            child nodes onto the stack.
            """
            stack, pointer = stack_and_pointer

            left_child = get_child_idx(level_order_idx, "l")
            right_child = get_child_idx(level_order_idx, "r")

            stack_updated = stack.at[pointer + 1].set(right_child)
            stack_updated = stack_updated.at[pointer + 2].set(left_child)

            pointer_updated = pointer + 2
            return stack_updated, pointer_updated

        def dont_push(stack_and_pointer: tuple) -> tuple:
            """If the node is a leaf, do nothing."""
            return stack_and_pointer

        # update the stack based on whether the current node is an operator
        new_stack, new_pointer = lax.cond(
            is_leaf,
            dont_push,
            push,
            operand=(stack, stack_pointer),
        )

        return (
            new_pre_order,
            new_pre_order_pointer,
            new_pre_level_map,
            new_stack,
            new_pointer,
        )

    def condition(state: tuple) -> ScalarBool:
        """Loop condition: Continue until stack is empty."""
        _, _, _, _, stack_pointer = state
        return stack_pointer >= 0

    # iterate over the sample using lax.while_loop
    final_pre_order, num_nodes, final_pre_level_map, _, _ = lax.while_loop(
        condition,
        traverse_tree,
        initial_state,
    )
    return final_pre_order, num_nodes, final_pre_level_map


@jit
def level_order_to_post_order_and_metrics(
    level_order: Int[jnp.ndarray, " D"],
    initial_state: tuple[
        Int[jnp.ndarray, " D"],
        ScalarInt,
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
        ScalarInt,
        ScalarInt,
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
    ],
    is_operator: Bool[jnp.ndarray, " D"],
    max_nodes: ScalarInt,
    max_depth: ScalarInt,
) -> tuple[
    Int[jnp.ndarray, " D"],
    ScalarInt,
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, " D"],
]:
    """
    Convert a level-order expression of a tree to a post-order expression, and
    other key metrics of the tree. The post-order expression is used to evaluate the
    tree kernel efficiently.

    Returns:
    - the post-order traversal expression of the tree
    - the number of nodes in the tree
    - a map from the post-order index to the level-order index
    - an array containing the sizes of the nodes in the tree
    - an array containing the heights of the nodes in the tree
    - a map from the leaf index to the level-order index


    NOTE: This function should be called via the get_post_order method of the
    TreeKernel.

    Parameters
    ----------
    level_order : Int[jnp.ndarray, " D"]
        The level-order expression of the tree.
    initial_state : tuple
        The initial state, containing allocated arrays for:
        - the post-order expression
        - the post-order pointer
        - the post_level_map
        - the stack
        - the stack pointer
        - the last processed index
        - the node sizes
        - the node heights
        - the leaf level map
    is_operator : Bool[jnp.ndarray, " D"]
        A boolean array indicating whether the node in the tree expression is an
        operator or not.
    max_nodes : ScalarInt
        The maximum number of nodes in the tree expression.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The post-order expression of the tree.
    ScalarInt
        The index of the last non-empty node in the tree (in the post-order expression).
    Int[jnp.ndarray, " D"]
        A map from the post-order index to the level-order index. The value at a given
        index in the post-order expression corresponds to the index of the same node in
        the level-order expression.
    Int[jnp.ndarray, " D"]
        An array containing the sizes of the nodes in the tree. The size of a node is
        the number of nodes in the subtree rooted at that node. The size of a given node
        is the value at its index in the level-order expression.
    Int[jnp.ndarray, " D"]
        An array containing the heights of the nodes in the tree. The height of a node
        is the length of the longest path from the node to a leaf node (i.e. the height
        of a leaf node is 0). The height of a given node is the value at its index in
        the level-order expression.
    Int[jnp.ndarray, " D"]
        A map from the leaf index to the level-order index. The value at a given index
        in the leaf level map corresponds to the index of the node with these parameters
        in the level-order expression.
        NOTE: The leaf index is decided based on the depth of the tree, and might not be
        in a simple order. See 'gallifrey.utils.tree_helper.get_parameter_leaf_idx' for
        more information.

    """

    def traverse_tree(state: tuple) -> tuple:
        """Traverse the tree expression in level-order and convert it to
        pre-order."""

        (
            _,
            _,
            _,
            stack,
            stack_pointer,
            last_processed_idx,
            _,
            _,
            _,
        ) = state

        # pop the top of the stack
        level_order_idx = stack[stack_pointer]

        # check if the current node is a leaf
        right_child_idx = get_child_idx(level_order_idx, "r")
        is_leaf = jnp.logical_or(
            jnp.logical_or(
                right_child_idx > max_nodes - 1,
                level_order[right_child_idx - 1] == -1,
            ),
            right_child_idx == last_processed_idx,
        )

        def process_non_leaf(state: tuple) -> tuple:
            """If the node is not a leaf, push the node back onto the stack
            and push the children onto the stack.
            """

            (
                post_order,
                post_order_pointer,
                post_level_map,
                stack,
                stack_pointer,
                last_processed_idx,
                node_sizes,
                node_heights,
                leaf_level_map,
            ) = state

            left_child = get_child_idx(level_order_idx, "l")
            right_child = get_child_idx(level_order_idx, "r")

            # Push current node, then right child, then left child onto stack
            stack_updated = stack.at[stack_pointer].set(level_order_idx)
            stack_updated = stack_updated.at[stack_pointer + 1].set(right_child)
            stack_updated = stack_updated.at[stack_pointer + 2].set(left_child)
            stack_pointer_updated = stack_pointer + 2

            return (
                post_order,
                post_order_pointer,
                post_level_map,
                stack_updated,
                stack_pointer_updated,
                last_processed_idx,
                node_sizes,
                node_heights,
                leaf_level_map,
            )

        def process_leaf(state: tuple) -> tuple:
            """If the node is a leaf, add the node to the post-order output."""
            (
                post_order,
                post_order_pointer,
                post_level_map,
                stack,
                stack_pointer,
                _,
                node_sizes,
                node_heights,
                leaf_level_map,
            ) = state

            # add the current node to the post-order output
            post_order_updated = post_order.at[post_order_pointer].set(
                level_order[level_order_idx]
            )
            post_order_pointer_updated = post_order_pointer + 1

            # update the post_level_map
            post_level_map_updated = post_level_map.at[post_order_pointer].set(
                level_order_idx
            )

            # update stack pointer
            stack_pointer_updated = stack_pointer - 1

            # update last processed index
            last_processed_idx_updated = level_order_idx

            # calculate and store the node size and heights
            left_child_idx = get_child_idx(level_order_idx, "l")
            right_child_idx = get_child_idx(level_order_idx, "r")

            left_child_size, left_child_height, right_child_size, right_child_height = (
                lax.cond(
                    right_child_idx < max_nodes,
                    lambda: (
                        node_sizes[left_child_idx],
                        node_heights[left_child_idx],
                        node_sizes[right_child_idx],
                        node_heights[right_child_idx],
                    ),
                    lambda: (0, 0, 0, 0),
                )
            )

            bigger_height = jnp.where(
                left_child_height > right_child_height,
                left_child_height,
                right_child_height,
            )

            node_sizes_updated = node_sizes.at[level_order_idx].set(
                1 + left_child_size + right_child_size
            )
            node_heights_updated = node_heights.at[level_order_idx].set(
                1 + bigger_height
            )

            # update leaf level map if the current node is a kernel atom,
            # not an operator (this differs from being a leaf node, since
            # in scaffolds, operators can be leaf nodes)
            leaf_level_map_updated = jnp.where(
                is_operator[level_order[level_order_idx]],
                leaf_level_map,
                leaf_level_map.at[
                    get_parameter_leaf_idx(
                        level_order_idx,
                        max_depth,
                    )
                ].set(level_order_idx),
            )

            return (
                post_order_updated,
                post_order_pointer_updated,
                post_level_map_updated,
                stack,
                stack_pointer_updated,
                last_processed_idx_updated,
                node_sizes_updated,
                node_heights_updated,
                leaf_level_map_updated,
            )

        # update the stack based on whether the current node is an operator
        new_state = lax.cond(
            is_leaf,
            process_leaf,
            process_non_leaf,
            operand=state,
        )

        return new_state

    def condition(state: tuple) -> ScalarBool:
        """Loop condition: Continue until stack is empty."""
        _, _, _, _, stack_pointer, _, _, _, _ = state
        return stack_pointer >= 0

    # iterate using lax.while_loop
    (
        final_pre_order,
        num_nodes,
        final_post_level_map,
        _,
        _,
        _,
        final_node_sizes,
        final_node_heights,
        final_leaf_level_map,
    ) = lax.while_loop(
        condition,
        traverse_tree,
        initial_state,
    )

    # to get the longest path from the root to a leaf, subtract 1 from the height,
    # since we counted to total number of nodes from the root to the leaf, not the
    # number of edges
    final_node_heights -= 1

    return (
        final_pre_order,
        num_nodes,
        final_post_level_map,
        final_node_sizes,
        final_node_heights,
        final_leaf_level_map,
    )


@partial(
    jit,
    static_argnames=(
        "atoms",
        "operators",
        "data_shape",
        "num_atoms",
        "max_depth",
    ),
)
def _evaluate_tree(
    x: Float[jnp.ndarray, " "] | Float[jnp.ndarray, " D"] | Float[jnp.ndarray, "D 1"],
    y: Float[jnp.ndarray, " "] | Float[jnp.ndarray, " D"] | Float[jnp.ndarray, "D 1"],
    post_order_expression: Int[jnp.ndarray, " D"],
    post_level_map: Int[jnp.ndarray, " D"],
    is_operator: Bool[jnp.ndarray, " D"],
    parameters: Float[jnp.ndarray, " M N"],
    atoms: tuple[AbstractAtom | tp.Callable, ...],
    operators: tuple[AbstractOperator, ...],
    data_shape: tuple[int, int],
    num_atoms: int,
    max_depth: int,
) -> Float[jnp.ndarray, "..."]:
    """
    Evaluate a tree expression by traversing the tree in post-order
    and applying the kernel atom functions and operators.

    NOTE: We are using Equinox's bounded_while_loop to evaluate the tree
    expression. More importantly, we are using its buffers feature to
    store the stack in the loop. This is a workaround to some weird JAX
    behaviour with inplace updates in while loops.
    (THIS IS ACTUALLY NOT CURRENTLY USED ANYMORE, BUT WE LEAVE THIS HERE FOR
    REFERENCE)
    For more information, see
    https://github.com/jax-ml/jax/issues/10197#issuecomment-1621416280
    https://github.com/jax-ml/jax/issues/17640
    https://github.com/patrick-kidger/equinox/blob/main/equinox/internal/_loop/loop.py


    Parameters
    ----------
    x : Float[jnp.ndarray, " "] | Float[jnp.ndarray, " D"] | Float[jnp.ndarray, "D 1"]
        Input x data (0D or 1D array of shape (D, ) filled with floats).
    y : Float[jnp.ndarray, " "] | Float[jnp.ndarray, " D"] | Float[jnp.ndarray, "D 1"]
        Input y data (0D or 1D array of shape (D, ) filled with floats).
    post_order_expression : Int[jnp.ndarray, " D"]
        The post-order expression of the tree.
    post_level_map : Int[jnp.ndarray, " D"]
        The map from post order index to level order index.
    is_operator : Bool[jnp.ndarray, " D"]
        A boolean array indicating whether a item in the atom library is an operator.
    parameters : Float[jnp.ndarray, " M N"]
        A jnp array containing the parameters of the kernel functions in the tree
        kernel. The shape of the array is (M, N), where M is the number of nodes in
        the tree and N is the maximum number of parameters of the kernel functions.
    atoms : tuple[AbstractAtom, ...] | tuple[tp.Callable, ...]
        A tuple of kernel atom functions.
    operators : tuple[AbstractOperator, ...]
        A tuple of kernel operators.
    data_shape : tuple[int, int]
        The shape of the input data (len(x), len(y)).
    num_atoms : int
        The number of atoms in the atom library.
    max_depth : int
        The maximum depth of the tree.

    Returns
    -------
    Float[jnp.ndarray, "..."]
        The result of evaluating the tree expression.
    """

    def evaluate(state: tuple) -> tuple:
        """Evaluate the tree expression at a given node index."""
        current_idx, *_ = state

        tree_level_idx = post_level_map[current_idx]
        node_value = post_order_expression[current_idx]
        is_op = is_operator[node_value]

        def eval_leaf_node(state: tuple) -> tuple:
            """Evaluate a leaf node."""
            current_idx, stack, pointer = state
            kernel_evaluation = lax.switch(
                node_value,
                atoms,
                x,
                y,
                parameters[get_parameter_leaf_idx(tree_level_idx, max_depth)],
            )
            new_stack = stack.at[pointer].set(kernel_evaluation)
            return (current_idx + 1, new_stack, pointer + 1)

        def eval_operator_node(state: tuple) -> tuple:
            """Evaluate an operator node."""
            _, stack, pointer = state
            left_child = jnp.array([stack[pointer - 1]])
            right_child = jnp.array([stack[pointer - 2]])
            kernel_evaluation = lax.switch(
                node_value - num_atoms,
                operators,
                left_child,
                right_child,
            )
            new_stack = stack.at[pointer - 2].set(kernel_evaluation)
            return (current_idx + 1, new_stack, pointer - 1)

        new_state = lax.cond(
            is_op,
            eval_operator_node,
            eval_leaf_node,
            operand=state,
        )
        return new_state

    def condition(state: tuple) -> Bool[jnp.ndarray, " "]:
        """Loop condition: Continue until the end of the tree expression (by
        encountering empty node, which is given by -1 value in
        post_order_expression).
        """
        current_idx, *_ = state
        return post_order_expression[current_idx] >= 0

    # def buffers(state: tuple) -> tp.Any:
    #     """Turn stack into Equinox buffer."""
    #     return state[1]

    # create initital state
    initial_stack = jnp.zeros(
        [
            calculate_max_stack_size(max_depth),
            data_shape[0],
            data_shape[1],
        ]
    )
    initial_pointer = 0
    initial_index = 0
    initital_state = (initial_index, initial_stack, initial_pointer)

    _, final_stack, *_ = bounded_while_loop(
        condition,
        evaluate,
        initital_state,
        max_steps=int(calculate_max_nodes(max_depth)),
        kind="checkpointed",
        # buffers=buffers,
        checkpoints=5,  # using checkpoint equation in equinox and assuming,
        # kernels with only ever have up to ~20 nodes,
        # might need to be adjusted for larger trees
    )
    return final_stack[0]


def tree_visualization(
    pre_order_expression: Int[jnp.ndarray, " D"],
    pre_level_map: Int[jnp.ndarray, " D"],
    atom_library: tuple[AbstractAtom | AbstractOperator, ...],
    parameters: Float[jnp.ndarray, " M N"],
    is_operator: Bool[jnp.ndarray, " D"],
    root_idx: Int[jnp.ndarray, ""],
    max_depth: ScalarInt,
    include_parameters: bool = True,
    print_str: bool = True,
    return_str: bool = False,
) -> None | str:
    """
    Construct a visual representation of a tree expression.

    Parameters
    ----------
    pre_order_expression : Int[jnp.ndarray, " D"]
        The tree expression in pre-order traversal notation.
    pre_level_map : Int[jnp.ndarray, " D"]
        A map from the pre-order index to the level-order index. The value at a given
        index in the pre-order expression corresponds to the index of the same node in
        the level-order expression.
    atom_library : tuple[AbstractAtom | AbstractOperator, ...]
        A tuple of kernel atom functions and operators to evaluate the tree expression.
    parameters : Float[jnp.ndarray, " M N"]
        The parameters of the kernel functions in the tree kernel.
    is_operator : Bool[jnp.ndarray, " D"]
        A boolean array indicating whether a item in the atom library is an operator.
    root_idx : Int[jnp.ndarray, ""]
        The index of the root node in the tree expression.
    max_depth : ScalarInt
        The maximum depth of the tree.
    include_parameters : bool, optional
        Whether to include the kernel parameters in the string, by default True.
    print_str : bool, optional
        Whether to print the visual representation of the tree, by default True.
    return_str : bool, optional
        Whether to return the visual representation as a string, by default False

    Returns
    -------
    str
        A string containing the visual representation of the tree expression (if
        return_str is True).
    """
    descr_str = ""

    for pre_order_idx, node_value in enumerate(pre_order_expression):
        tree_level_idx = pre_level_map[pre_order_idx]
        atom = atom_library[node_value]

        depth = get_depth(tree_level_idx) - get_depth(root_idx)
        prefix = "    " * (depth - 1) + "└── " if depth > 0 else ""

        node_representation = f"{atom.name}"

        if not is_operator[node_value] and include_parameters:
            node_parameters = parameters[
                get_parameter_leaf_idx(tree_level_idx, max_depth)
            ]
            parameter = node_parameters[: atom.num_parameter]
            node_representation += f": {parameter}"

        descr_str += f"{prefix}{node_representation}\n"

    if print_str:
        print(descr_str)
    if return_str:
        return descr_str
    return None



from functools import partial

import beartype.typing as tp
import jax
import jax.random as jr
from flax import nnx
from jax import jit
from jaxtyping import PRNGKeyArray
from tensorflow_probability.substrates.jax.distributions import Distribution

from gallifrey.data import Dataset
from gallifrey.inference.hmc import get_hmc_objective, hmc_inference_loop
from gallifrey.kernels.prior import KernelPrior
from gallifrey.parameter import ParticleParameter
from gallifrey.particles import transform_particle_parameters
from gallifrey.utils.typing import ScalarInt


@partial(
    jit,
    static_argnames=(
        "data",
        "kernel_prior",
        "noise_prior",
        "n_hmc",
        "hmc_sampler_factory",
    ),
)
def rejuvenate_particle_parameters(
    key: PRNGKeyArray,
    particle_state: nnx.State,
    data: Dataset,
    kernel_prior: KernelPrior,
    noise_prior: Distribution,
    n_hmc: ScalarInt,
    hmc_sampler_factory: tp.Callable,
) -> tuple[nnx.State, ScalarInt]:
    """
    Rejuvenate GP parameters using HMC.

    Parameters
    ----------
    key : PRNGKeyArray
        The random key.
    particle_state : nnx.State
        The current particle state.
    data : Dataset
        The dataset object containing
        the input x and output y.
    kernel_prior : KernelPrior
        The kernel prior object.
    noise_prior : Distribution
        The noise prior distribution.
    n_hmc : ScalarInt
        The number of HMC steps.
    hmc_sampler_factory : tp.Callable
        The HMC sampler factory function,
        which takes an objective function and
        returns a Blackjax HMC sampler.

    Returns
    -------
    nnx.State
        The particle state with rejuvenated
        parameters.
    ScalarInt
        The number of accepted HMC steps.

    """

    # get HMC objective and sampler
    hmc_objective = get_hmc_objective(
        particle_state,
        data,
        kernel_prior,
        noise_prior,
    )
    hmc_sampler = hmc_sampler_factory(hmc_objective)

    # transform particle parameters to unconstrained space and
    # create initial HMC state
    unconstrained_particle_state = transform_particle_parameters(
        particle_state,
        kernel_prior,
        inverse=True,
    )
    unconstrained_parameter_state, *static_state = unconstrained_particle_state.split(
        ParticleParameter, ...
    )
    initial_hmc_state = hmc_sampler.init(unconstrained_parameter_state)

    # run the HMC inference loop
    key, sample_key = jr.split(key)
    final_parameter_state, other = hmc_inference_loop(
        sample_key,
        jax.jit(hmc_sampler.step),
        initial_hmc_state,
        n_hmc,
    )
    history, info = other
    # HMCInfo contains (maybe we could return more):
    # - momentum
    # - acceptance_rate
    # - is_accepted
    # - is_divergent
    # - energy
    # - proposal
    # - step_size
    # - num_integration_steps

    # put particle state back together and transform to constrained space
    new_particle_state = nnx.State.merge(
        final_parameter_state.position,  # type: ignore
        *static_state,
    )
    constrained_particle_state = transform_particle_parameters(
        new_particle_state,
        kernel_prior,
        inverse=False,
    )

    return constrained_particle_state, info.is_accepted.sum()



from functools import partial

import beartype.typing as tp
import jax.random as jr
from flax import nnx
from jax import jit, lax
from jax import numpy as jnp
from jaxtyping import Int, PRNGKeyArray
from tensorflow_probability.substrates.jax.distributions import Distribution

from gallifrey.data import Dataset
from gallifrey.inference.parameter_rejuvenation import rejuvenate_particle_parameters
from gallifrey.kernels.prior import KernelPrior
from gallifrey.moves.particle_move import structure_move
from gallifrey.utils.typing import ScalarBool, ScalarInt


@partial(
    jit,
    static_argnames=(
        "data",
        "kernel_prior",
        "noise_prior",
        "n_mcmc",
        "n_hmc",
        "fix_noise",
        "hmc_sampler_factory",
        "verbosity",
    ),
)
def rejuvenate_particle(
    key: PRNGKeyArray,
    particle_state: nnx.State,
    data: Dataset,
    kernel_prior: KernelPrior,
    noise_prior: Distribution,
    n_mcmc: ScalarInt,
    n_hmc: ScalarInt,
    fix_noise: bool,
    hmc_sampler_factory: tp.Callable,
    verbosity: int = 0,
) -> tuple[nnx.State, nnx.State, ScalarInt, ScalarInt]:
    """
    Rejuvenates a particle's state using a structure MCMC move followed by
    parameter HMC moves.

    This function performs a sequence of rejuvenation steps for a
    single particle state. It iterates `n_mcmc` times, applying the
    `rejuvenation_step` function in each iteration.

    Parameters
    ----------
    key : PRNGKeyArray
        Random key for the MCMC sampling.
    particle_state : nnx.State
        The current state of the particle to be rejuvenated.
    data : Dataset
        The dataset object containing the input and output data.
    kernel_prior : KernelPrior
        The kernel prior for the kernel structure.
    noise_prior : Distribution
        The prior distribution for the noise variance.
    n_mcmc : ScalarInt
        The number of MCMC iterations (rejuvenation steps) to perform.
    n_hmc : ScalarInt
        The number of Hamiltonian Monte Carlo (HMC) steps to perform within each
        parameter rejuvenation step (inside `rejuvenation_step`).
    fix_noise : bool
        A boolean flag indicating whether the noise variance is fixed or trainable.
    hmc_sampler_factory : tp.Callable
        A factory function that creates an HMC sampler.
    verbosity : int
        The verbosity level. Debug information is printed if `verbosity > 0`.

    Returns
    -------
    nnx.State
        The final rejuvenated particle state after `n_mcmc` iterations.
    nnx.State
        The history of the rejuvenation steps after every MCMC iteration.
    ScalarInt
        The total number of MCMC steps accepted during the rejuvenation process.
    ScalarInt
        The total number of HMC steps accepted during the rejuvenation process.
    """

    def scan_func(
        loop_state: tuple[nnx.State, Int[jnp.ndarray, ""], Int[jnp.ndarray, ""]],
        key: PRNGKeyArray,
    ) -> tuple[tuple[nnx.State, Int[jnp.ndarray, ""], Int[jnp.ndarray, ""]], nnx.State]:
        """
        A wrapper function for a single `rejuvenation_step` to be
        passed to the scan function.
        """
        particle_state, accepted_mcmc, accepted_hmc = loop_state

        new_state, mcmc_accepted, n_hmc_accepted = rejuvenation_step(
            key,
            particle_state,
            data,
            kernel_prior,
            noise_prior,
            n_hmc,
            fix_noise,
            hmc_sampler_factory,
            verbosity,
        )
        accepted_mcmc = accepted_mcmc + mcmc_accepted
        accepted_hmc = accepted_hmc + n_hmc_accepted

        return (new_state, accepted_mcmc, accepted_hmc), new_state

    final_loop_state, history = lax.scan(
        scan_func,
        (particle_state, jnp.array(0), jnp.array(0)),
        jr.split(key, int(n_mcmc)),
    )

    final_state, accepted_mcmc, accepted_hmc = final_loop_state

    return final_state, history, accepted_mcmc, accepted_hmc


@partial(
    jit,
    static_argnames=(
        "data",
        "kernel_prior",
        "noise_prior",
        "n_hmc",
        "fix_noise",
        "hmc_sampler_factory",
        "verbosity",
    ),
)
def rejuvenation_step(
    key: PRNGKeyArray,
    particle_state: nnx.State,
    data: Dataset,
    kernel_prior: KernelPrior,
    noise_prior: Distribution,
    n_hmc: ScalarInt,
    fix_noise: bool,
    hmc_sampler_factory: tp.Callable,
    verbosity: int = 0,
) -> tuple[nnx.State, ScalarBool, ScalarInt]:
    """
    Performs a single rejuvenation step for a
    article state.

    This function implements a single step of the rejuvenation process.
    It consists of two main moves:
        1. Structure Move: Proposes a new kernel structure based on the `kernel_prior`.
        2. Parameter Rejuvenation: If the structure move is accepted, rejuvenates
        the kernel parameters and potentially the noise variance using
        Hamiltonian Monte Carlo (HMC).

    Parameters
    ----------
    key : PRNGKeyArray
        Random key for the rejuvenation step.
    particle_state : nnx.State
        The current state of the particle to be rejuvenated.
    data : Dataset
        The dataset object containing the input and output data.
    kernel_prior : KernelPrior
        The kernel prior for the kernel structure.
    noise_prior : Distribution
        The prior distribution for the noise variance.
    n_hmc : ScalarInt
        The number of Hamiltonian Monte Carlo (HMC) steps to perform during
        parameter rejuvenation.
    fix_noise : bool
        A boolean flag indicating whether the noise variance is fixed or trainable.
    hmc_sampler_factory : tp.Callable
        A factory function that creates an HMC sampler.
    verbosity : int
        The verbosity level. Debug information of the structure move
        is printed if `verbosity > 1`.

    Returns
    -------
    nnx.State
        The rejuvenated particle state after performing the structure move
        and (potentially) parameter rejuvenation.
    ScalarBool
        A boolean flag indicating whether the structure move was accepted.
    ScalarInt
        The number of HMC steps accepted during the parameter rejuvenation.
    """

    key, structure_key, hmc_key = jr.split(key, 3)

    new_particle_state, accepted = structure_move(
        structure_key,
        particle_state,
        kernel_prior,
        noise_prior,
        data,
        fix_noise=fix_noise,
        verbosity=verbosity,
    )

    new_particle_state, n_hmc_accepted = lax.cond(
        accepted,
        lambda key: rejuvenate_particle_parameters(
            key,
            new_particle_state,
            data,
            kernel_prior,
            noise_prior,
            n_hmc,
            hmc_sampler_factory,
        ),
        lambda key: (new_particle_state, jnp.array(0)),
        hmc_key,
    )

    return new_particle_state, jnp.array(accepted, dtype=jnp.bool), n_hmc_accepted



from functools import partial

from flax import nnx
from jax import debug, jit, lax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Int, PRNGKeyArray

from gallifrey.kernels.prior import KernelPrior, TreeStructurePrior
from gallifrey.utils.tree_helper import get_child_idx, pick_random_node
from gallifrey.utils.typing import ScalarBool, ScalarInt


@partial(jit, static_argnames=("kernel_prior", "verbosity"))
def subtree_replace_move(
    key: PRNGKeyArray,
    kernel_state: nnx.State,
    kernel_prior: KernelPrior,
    verbosity: int = 0,
) -> nnx.State:
    """
    Perform a subtree replacement move on a kernel state.

    In this move, a subtree is replaced with a new
    subtree sampled from the prior. The root of the subtree
    to be replaced is chosen uniformly at random.

    The parameter from the old tree are copied to the new tree,
    while the parameters for the new subtree are sampled from the prior.

    Parameters
    ----------
    key : PRNGKeyArray
        The random key.
    kernel_state : nnx.State
        The original kernel state, must have attributes:
        - tree_expression
        - post_level_map
    kernel_prior : KernelPrior
        The prior to sample the new subtree from.
    verbosity : int, optional
        The verbosity level, by default 0. Debug information is printed
        if verbosity > 2.


    Returns
    -------
    nnx.State
        The updated kernel state after the move.

    """
    key, structure_key, parameter_key = jr.split(key, 3)

    # perform the subtree replacement move on the structure
    new_structure, changes, _ = structure_subtree_replace_move(
        structure_key,
        kernel_state.tree_expression.value,  # type: ignore
        kernel_state.post_level_map.value,  # type: ignore
        kernel_prior.kernel_structure_prior,
        verbosity=verbosity,
    )

    # create a new kernel with the updated structure
    new_kernel = nnx.merge(kernel_prior.graphdef, kernel_state)
    new_kernel.init_tree(
        new_structure,
        kernel_prior.kernel_library,
        kernel_prior.max_depth,
        kernel_prior.num_datapoints,
    )

    # sample the new parameters, while keeping the old ones
    _, blank_new_state = nnx.split(new_kernel)
    new_state, _ = kernel_prior.parameter_prior.sample_subset(
        parameter_key,
        blank_new_state,
        considered_nodes=changes,
    )
    return new_state


def structure_subtree_replace_move(
    key: PRNGKeyArray,
    tree_expression: Int[jnp.ndarray, " D"],
    post_level_map: Int[jnp.ndarray, " D"],
    tree_prior: TreeStructurePrior,
    verbosity: int = 0,
) -> tuple[
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, " D"],
    ScalarInt,
]:
    """
    Perform a subtree replacement move on the kernel structure array.
    In this move, a subtree is replaced with a new
    subtree sampled from the prior. The root of the subtree
    to be replaced is chosen uniformly at random.

    Parameters
    ----------
    key : PRNGKeyArray
        The random key.
    tree_expression : Int[jnp.ndarray, " D"]
        The tree expression (level order).
    post_level_map : Int[jnp.ndarray, " D"]
        The array that maps the post order index to the level order index.
        (This is useful, since the post_level_map contains all the indices
        of the nodes in the tree).
    tree_prior : TreeStructurePrior
        The prior to sample the new subtree from.
    verbosity : int, optional
        The verbosity level, by default 0. Debug information is printed
        if verbosity > 2.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The updated tree expression after the move.
    Int[jnp.ndarray, " D"]
        An array containing the indices of the nodes that were changed.
    ScalarInt
        The index where the new subtree was attached.

    """
    key, idx_key, subtree_key = jr.split(key, 3)

    # sample a node to replace uniformly from non-empty nodes
    replace_idx = pick_random_node(idx_key, post_level_map)

    # if tree_expression[replace_idx] == empty_node_value:
    #     raise RuntimeError(
    #         "The chosen index is that of an empty node, "
    #         "that should not be possible. Check inputs."
    #     )

    if verbosity > 2:
        debug.print("Replacing node at index {}", replace_idx)

    updated_tree_expression, changed_nodes = replace_subtree_structure(
        subtree_key,
        tree_expression,
        replace_idx,
        tree_prior,
        verbosity=verbosity,
    )
    return updated_tree_expression, changed_nodes, replace_idx


def replace_subtree_structure(
    key: PRNGKeyArray,
    tree_expression: Int[jnp.ndarray, " D"],
    replace_idx: ScalarInt,
    tree_prior: TreeStructurePrior,
    verbosity: int = 0,
) -> tuple[
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, " D"],
]:
    """
    Replace a subtree in the tree expression with a
    new subtree sampled from the prior.

    In contrast to the subtree_replace_move function,
    this function uses a fixed index for
    the root of the subtree to replace.

    Parameters
    ----------
    key : PRNGKeyArray
        The random key.
    tree_expression : Int[jnp.ndarray, " D"]
        The tree expression (level order).
    replace_idx : ScalarInt
        The index of the node where to attach the new subtree.
    tree_prior : TreeStructurePrior
        The prior to sample the new subtree from.
    verbosity : int, optional
        The verbosity level, by default 0. Debug information is printed
        if verbosity > 2.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The updated tree expression.
    Int[jnp.ndarray, " D"]
        An array containing the indices of the nodes that were changed.
    """

    new_subtree = tree_prior.sample_single(
        key,
        root_idx=replace_idx,
    )

    if verbosity > 2:
        debug.print("New subtree: {}", new_subtree)

    return replace_subtree_structure_with(
        tree_expression,
        new_subtree,
        replace_idx,
    )


def replace_subtree_structure_with(
    tree_expression: Int[jnp.ndarray, " D"],
    subtree_expression: Int[jnp.ndarray, " D"],
    replace_idx: ScalarInt,
) -> tuple[
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, " D"],
]:
    """
    Replace a subtree in the tree_expression with
    another subtree. The subtree to replace is specified
    by the replace_idx.

    In contrast to the subtree_replace_move and replace_subtree functions,
    this function takes the tree_expression and subtree_expression as inputs.
    No random key is required.

    Parameters
    ----------
    tree_expression : Int[jnp.ndarray, " D"]
        The tree expression (level order).
    subtree_expression : Int[jnp.ndarray, " D"]
        The subtree expression (level order). The array must
        have the same length as the tree_expression.
    replace_idx : ScalarInt
        The index of the node where to attach the new subtree.


    Returns
    -------
    Int[jnp.ndarray, " D"]
        The updated tree expression, with the subtree replaced.
    Int[jnp.ndarray, " D"]
        An array containing the indices of the nodes that were changed.

    """

    max_nodes = tree_expression.size

    # if max_nodes != subtree_expression.size:
    #     raise ValueError(
    #         "tree_expression and subtree_expression must have the same length."
    #     )

    # create initial state
    changed_values = jnp.full(max_nodes, -1, dtype=tree_expression.dtype)
    changed_pointer = 0
    stack = jnp.copy(tree_expression).at[0].set(replace_idx)
    stack_pointer = 0
    initial_state = (
        tree_expression,
        subtree_expression,
        changed_values,
        changed_pointer,
        stack,
        stack_pointer,
    )

    new_tree_expression, changed_values, _ = _replace_subtree_structure_with(
        initial_state,
        max_nodes,
    )

    return new_tree_expression, changed_values


@jit
def _replace_subtree_structure_with(
    initial_state: tuple[
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
        Int[jnp.ndarray, " D"],
        ScalarInt,
        Int[jnp.ndarray, " D"],
        ScalarInt,
    ],
    max_nodes: ScalarInt,
) -> tuple[
    Int[jnp.ndarray, " D"],
    Int[jnp.ndarray, " D"],
    ScalarInt,
]:
    """
    Replace a subtree in the tree expression with another subtree.

    NOTE: This function is not meant to be called directly. Use replace_subtree_with
    instead.

    Parameters
    ----------
    initial_state : tuple
        The initial state of the traversal. Contains the tree expression,
        the subtree expression, the changed values, the changed pointer,
        the stack and the stack pointer.
    max_nodes : ScalarInt
        The maximum number of nodes in the tree.

    Returns
    -------
    Int[jnp.ndarray, " D"]
        The updated tree expression with the subtree replaced.
    Int[jnp.ndarray, " D"]
        An array containing the indices of the nodes that were changed (same length
        as tree_expression).
    ScalarInt
        The number of nodes that were changed.


    """

    def traverse(state: tuple) -> tuple:
        """Traverse the tree expression in level-order."""
        (
            tree_expression,
            subtree_expression,
            changed_values,
            changed_pointer,
            stack,
            pointer,
        ) = state

        idx = stack[pointer]

        # check if the current node is a leaf in either tree
        right_child_idx = get_child_idx(idx, "r")
        is_leaf = jnp.logical_or(
            right_child_idx > max_nodes - 1,
            jnp.logical_and(
                tree_expression[right_child_idx] == -1,
                subtree_expression[right_child_idx] == -1,
            ),
        )

        def process_leaf(stack_and_pointer: tuple) -> tuple:
            """If the current node is a leaf, move parent."""
            stack, pointer = stack_and_pointer
            return stack, pointer - 1

        def process_non_leaf(stack_and_pointer: tuple) -> tuple:
            """If the current node is not a leaf, update the stack."""
            stack, pointer = stack_and_pointer

            left_child_idx = get_child_idx(idx, "l")
            right_child_idx = get_child_idx(idx, "r")

            new_stack = stack.at[pointer].set(left_child_idx)
            new_stack = new_stack.at[pointer + 1].set(right_child_idx)

            return new_stack, pointer + 1

        # update the stack and pointer
        new_stack, new_pointer = lax.cond(
            is_leaf,
            process_leaf,
            process_non_leaf,
            operand=(stack, pointer),
        )

        # override the current node in the tree expression with subtree node
        old_value = tree_expression[idx]
        new_value = subtree_expression[idx]
        tree_expression_updated = tree_expression.at[idx].set(new_value)

        # if values are different, update the changed values and pointer
        changed_values_updated, changed_pointer_updated = lax.cond(
            old_value != new_value,
            lambda x: (x[0].at[x[1]].set(idx), x[1] + 1),
            lambda x: x,
            operand=(changed_values, changed_pointer),
        )

        return (
            tree_expression_updated,
            subtree_expression,
            changed_values_updated,
            changed_pointer_updated,
            new_stack,
            new_pointer,
        )

    def condition(state: tuple) -> ScalarBool:
        """Check if the traversal is complete."""
        _, _, _, _, _, pointer = state
        return pointer >= 0

    # iterate over the sample using lax.while_loop
    new_tree_expression, _, changed_values, changed_pointer, _, _ = lax.while_loop(
        condition,
        traverse,
        initial_state,
    )

    return new_tree_expression, changed_values, changed_pointer



#########################
# Code adapted from GPJax
#########################
from copy import deepcopy

import jax.numpy as jnp
import tensorflow_probability.substrates.jax.bijectors as tfb
from flax import nnx
from jax import lax
from jax.scipy import linalg
from jaxtyping import Array, Float
from tensorflow_probability.substrates.jax.distributions import (
    Distribution,
    MultivariateNormalFullCovariance,
)

from gallifrey.data import Dataset
from gallifrey.kernels.prior import KernelPrior
from gallifrey.kernels.tree import TreeKernel
from gallifrey.parameter import NoiseParameter, transform_kernel_parameters
from gallifrey.utils.probability_calculations import calculate_covariance_matrix
from gallifrey.utils.typing import ScalarBool, ScalarFloat

SOFTPLUS_BIJECTOR = tfb.Softplus()


class Particle(nnx.Module):
    """
    The `Particle` class. This class is based on of the GPJax
    `ConjugatePosterior`.

    It takes a kernel (instance of `TreeKernel`), noise variance
    (float), and jitter (float) as input. The jitter is used to
    ensure that the covariance matrix is positive definite (avoiding
    numerical instabilities in case of small eigenvalues).

    In contrast to the `ConjugatePosterior` class, the `Particle` currently
    has more limited features, specifically we
        - do not support mean functions (that implies a zero mean function,
          in use cases the data should be appropriately centered).
        - do not support likelihoods other than Gaussian.

    Attributes
    ----------
    kernel : TreeKernel
        The kernel defining the covariance function of the Gaussian
        process. Must be a `TreeKernel` instance (from `gallifrey.kernels`).

    noise_variance : nnx.Variable | NoiseParameter
        The (observational) noise variance of the Gaussian process. Depending
        on the value of the `trainable_noise_variance` parameter during initialization,
        the noise variance will be either considered a trainable parameter (instance
        of `NoiseParameter`) or a fixed parameter (instance of `nnx.Variable`).

    jitter : ScalarFloat
        The jitter term to ensure numerical stability of the covariance
        matrix.
    """

    def __init__(
        self,
        kernel: TreeKernel,
        noise_variance: ScalarFloat,
        trainable_noise_variance: bool = False,
        jitter: float = 1e-6,
    ):
        """
        Initialize the particle.

        Parameters
        ----------
        kernel : TreeKernel
            The kernel defining the covariance function of the Gaussian
            process. Must be a `TreeKernel` instance (from `gallifrey.kernels`).
        noise_variance : ScalarFloat
            The (observational) noise variance of the Gaussian process.
        trainable_noise_variance : bool, optional
            Whether to treat the noise variance as a trainable parameter,
            by default False.
        jitter : float, optional
            The jitter term to ensure numerical stability of the covariance
            matrix, by default 1e-6.
        """
        self.kernel = kernel
        noise_variance = jnp.asarray(noise_variance)
        if trainable_noise_variance:
            self.noise_variance = NoiseParameter(noise_variance)  # type: ignore
        else:
            self.noise_variance = nnx.Variable(noise_variance)  # type: ignore

        self.jitter = jitter

    def __str__(self) -> str:
        """
        Print a string representation of the particle.

        Returns
        -------
        str
            A string representation of the particle.
        """
        variance = f"Noise Variance : {self.noise_variance.value}"
        kernel = self.kernel.__str__()

        return f"{variance}\n{kernel}"

    def display(self) -> None:
        """
        Display the particle.
        """
        print(self.__str__())

    def predictive_distribution(
        self,
        xpredict: Float[Array, " D"],
        data: Dataset,
        latent: bool = False,
    ) -> Distribution:
        """
        Calculate the predictive distribution of the Gaussian process
        for the particle.

        Inputs will be the points to predict and the training data that
        the GP gets conditioned on.

        The distribution returned will be a MultivariateNormalFullCovariance
        distribution from `tensorflow_probability.substrates.jax.distributions`, see
        https://www.tensorflow.org/probability/api_docs/python/tfp/distributions/MultivariateNormalFullCovariance

        If `latent` is True, the predictive distribution of the latent
        function is returned, i.e. the distribution of the function
        values without the observational noise. If False, the predictive
        distribution of the full data-generating model is returned, which
        includes the observational noise


        Parameters
        ----------
        xpredict : Float[Array, " D"]
            The points to predict, as a 1D array.
        data : Dataset
            The training data that the GP is conditioned on,
            must be a `Dataset` instance from `gallifrey.data`.
        latent : bool, optional
            Whether to return the predictive distribution of the latent
            function (without observational noise), by default False.

        Returns
        -------
        Distribution
            A tensorflow probability distribution object representing
            the predictive distribution of the Gaussian process. (Specifically,
            a MultivariateNormalFullCovariance distribution from
            `tensorflow_probability.substrates.jax.distributions`).
        """

        # unpack conditioning data
        xtrain, ytrain = data.x, data.y

        # the points to be predicted, as correct type
        t = jnp.asarray(xpredict)

        # calculate covariance matrix (Σ = Kxx + Io²) for conditioning data
        Kxx = self.kernel.gram(xtrain)
        Sigma = calculate_covariance_matrix(
            Kxx,
            self.noise_variance.value,
            self.jitter,
        )

        # calculate gram and cross_covariance for prediction points
        Ktt = self.kernel.gram(t)
        Kxt = self.kernel.cross_covariance(xtrain, t)

        # solve for Σ⁻¹Kxt | [len(xtrain),len(t)] using cholesky decomposition
        cho_fac = linalg.cho_factor(Sigma)
        Sigma_inv_Kxt = linalg.cho_solve(cho_fac, Kxt)

        # calculate predictive mean  Ktx (Kxx + Io²)⁻¹y (assumes zero mean function)
        predictive_mean = jnp.matmul(Sigma_inv_Kxt.T, ytrain)

        # calculate latent covariance function  Ktt  -  Ktx (Kxx + Io²)⁻¹ Kxt
        latent_covariance = Ktt - jnp.matmul(Kxt.T, Sigma_inv_Kxt)
        latent_covariance += jnp.eye(latent_covariance.shape[0]) * self.jitter

        # The covariance matrix that we've calculated so far, is the covariance
        # of the latent distribution as estimated by the Gaussian process. To
        # get the full covariance matrix of the model (for the data generating
        # process), we need to add the noise observational noise variance.
        if latent:
            predictive_covariance = latent_covariance
        else:
            # add noise variance to the diagonal of the covariance matrix
            predictive_covariance = (
                latent_covariance
                + jnp.eye(latent_covariance.shape[0]) * self.noise_variance
            )

        return MultivariateNormalFullCovariance(
            predictive_mean,
            predictive_covariance,
        )


def transform_particle_parameters(
    particle_state: nnx.State,
    kernel_prior: KernelPrior,
    inverse: ScalarBool = False,
) -> nnx.State:
    """
    Transform parameter of a particle state (kernel parameters
    and noise variance) based on the support mapping and bijectors.

    This function is primarily used to transform the parameters
    between a constrained and unconstrained space.

    Parameters
    ----------
    particle_state : nnx.State
        The original particle state.
    kernel_prior : KernelPrior
        The kernel prior that contains the support mapping and bijectors.
    inverse : ScalarBool, optional
        If True, the inverse transformation is applied, by default False.

    Returns
    -------
    nnx.State
        The particle state with transformed parameters.
    """
    # get parameters and state
    num_parameter_array = kernel_prior.parameter_prior.num_parameter_array
    max_leaves = kernel_prior.parameter_prior.max_leaves
    max_atom_parameters = kernel_prior.parameter_prior.max_atom_parameters
    support_mapping_array = kernel_prior.parameter_prior.support_mapping_array
    support_bijectors = kernel_prior.support_bijectors

    kernel_state = particle_state.kernel
    noise_variance = jnp.array(particle_state.noise_variance.value)  # type: ignore

    # transform the noise standard deviation (using softplus bijector)
    transformed_noise_variance = lax.cond(
        inverse,
        SOFTPLUS_BIJECTOR.inverse,
        SOFTPLUS_BIJECTOR.forward,
        noise_variance,
    )

    # transform the kernel parameters
    transformed_kernel_state = transform_kernel_parameters(
        kernel_state,
        num_parameter_array,
        max_leaves,
        max_atom_parameters,
        support_mapping_array,
        support_bijectors,
        inverse,
    )

    # create a new state with the transformed parameters
    new_state = deepcopy(particle_state)
    new_state.kernel = transformed_kernel_state  # type: ignore
    new_state.noise_variance = (  # type: ignore
        particle_state.noise_variance.replace(
            transformed_noise_variance,
        )  # type: ignore
    )

    return new_state



from gallifrey.kernels.atoms import (
    ConstantAtom,
    LinearAtom,
    LinearWithShiftAtom,
    Matern12Atom,
    Matern32Atom,
    Matern52Atom,
    PeriodicAtom,
    PoweredExponentialAtom,
    ProductOperator,
    RationalQuadraticAtom,
    RBFAtom,
    SumOperator,
    WhiteAtom,
)
from gallifrey.kernels.library import KernelLibrary
from gallifrey.kernels.prior import KernelPrior, ParameterPrior, TreeStructurePrior
from gallifrey.kernels.tree import TreeKernel

__all__ = [
    "ConstantAtom",
    "LinearAtom",
    "LinearWithShiftAtom",
    "Matern12Atom",
    "Matern32Atom",
    "Matern52Atom",
    "PeriodicAtom",
    "PoweredExponentialAtom",
    "ProductOperator",
    "RationalQuadraticAtom",
    "RBFAtom",
    "SumOperator",
    "WhiteAtom",
    "KernelLibrary",
    "KernelPrior",
    "ParameterPrior",
    "TreeStructurePrior",
    "TreeKernel",
]



from functools import partial

import jax.numpy as jnp
from jax import jit
from jax.scipy import linalg
from jaxtyping import Float

from gallifrey.data import Dataset
from gallifrey.utils.typing import ScalarFloat


# @partial(jit, static_argnames=("jitter", "data"))
def calculate_marginal_log_likelihood(
    kernel_gram: Float[jnp.ndarray, "D D"],
    noise_variance: ScalarFloat,
    data: Dataset,
    jitter: float = 1e-6,
) -> ScalarFloat:
    """
    Calculate the log marginal likelihood p(y) for the
    state (k, theta, y, noise).

    Parameters
    ----------
    kernel_gram :  Float[jnp.ndarray, "D D"]
        The kernel gram matrix.
    noise_variance : ScalarFloat
        The noise variance.
    data : Dataset
        A dataset object containing the input x and output y.
    jitter : float, optional
        The jitter term to add to the kernel gram matrix, by default 1e-6.
        (This is to ensure the matrix is positive definite.)

    Returns
    -------
    ScalarFloat
        The log marginal likelihood p(y) for the current state.

    """

    sigma_logdet, sigma_inv_r = calculate_sigma_logdet_and_inv_sigma_r(
        data.y,
        kernel_gram,
        noise_variance,
        jitter,
    )
    # compute marginal log likelihood, -1/2[ n log(2π) + log|Σ| + ( rᵀ Σ⁻¹ r ]
    log_prob = -0.5 * (
        data.n * jnp.log(2.0 * jnp.pi) + sigma_logdet + data.y.T @ sigma_inv_r
    )
    return log_prob


@partial(jit, static_argnames=("jitter"))
def calculate_sigma_logdet_and_inv_sigma_r(
    r: Float[jnp.ndarray, " D"],
    kernel_gram: Float[jnp.ndarray, " D D"],
    noise_variance: ScalarFloat,
    jitter: float = 1e-6,
) -> tuple[ScalarFloat, Float[jnp.ndarray, " D"]]:
    """
    Calculate the log determinant of the covariance matrix Σ for the
    Gaussian process, and solve for Σ⁻¹r.

    Parameters
    ----------
    r :  Float[jnp.ndarray, " D"]
        The input observational residuals
        (observations y - deterministic mean function).
    kernel_gram : Float[jnp.ndarray, "D D"]
        The kernel gram matrix.
    noise_variance : ScalarFloat
        The noise variance.
    jitter : float, optional
        The jitter term to add to the kernel gram matrix, by default 1e-6.
        (This is to ensure the matrix is positive definite.)

    Returns
    -------
    ScalarFloat
        The log determinant of the covariance matrix Σ.
    Float[jnp.ndarray, " D"]
        The solution for Σ⁻¹r.
    """

    # calculate Sigma matrix for GP, Σ = (Kxx + Io²) | [n,n]
    sigma_matrix = calculate_covariance_matrix(
        kernel_gram,
        noise_variance,
        jitter,
    )

    # solve for Σ⁻¹r | [n,1] using cholesky decomposition,
    # where Σ⁻¹r = L⁻ᵀ L⁻¹ r, with L = cholesky(Σ)
    cho_fac = linalg.cho_factor(sigma_matrix)
    sigma_inv_r = linalg.cho_solve(cho_fac, r)

    # calculate log determinant | [1,1]
    # log(|Σ|) =  log(|L|**2) = 2 log(|L|), with L = cholesky(Σ)
    sigma_logdet = jnp.sum(jnp.log(jnp.diag(cho_fac[0]))) * 2.0
    return sigma_logdet, sigma_inv_r


@partial(jit, static_argnames=("jitter"))
def calculate_inv_sigma_r(
    r: Float[jnp.ndarray, " D"],
    kernel_gram: Float[jnp.ndarray, " D D"],
    noise_variance: ScalarFloat,
    jitter: float = 1e-6,
) -> Float[jnp.ndarray, " D"]:
    """
    Calculate the covariance matrix Σ for the
    Gaussian process, and solve for Σ⁻¹r.

    Parameters
    ----------
    r :  Float[jnp.ndarray, " D"]
        The input observational residuals
        (observations y - deterministic mean function).
    kernel_gram : Float[jnp.ndarray, "D D"]
        The kernel gram matrix.
    noise_variance : ScalarFloat
        The noise variance.
    jitter : float, optional
        The jitter term to add to the kernel gram matrix, by default 1e-6.
        (This is to ensure the matrix is positive definite.)

    Returns
    -------
     Float[jnp.ndarray, " D"]
        The solution for Σ⁻¹r.
    """

    # calculate Sigma matrix for GP, Σ = (Kxx + Io²) | [n,n]
    sigma_matrix = calculate_covariance_matrix(
        kernel_gram,
        noise_variance,
        jitter,
    )
    # solve for Σ⁻¹r | [n,1] using cholesky decomposition,
    # where Σ⁻¹r = L⁻ᵀ L⁻¹ r, with L = cholesky(Σ)
    cho_fac = linalg.cho_factor(sigma_matrix)
    sigma_inv_r = linalg.cho_solve(cho_fac, r)
    return sigma_inv_r


@partial(jit, static_argnames=("jitter"))
def calculate_covariance_matrix(
    kernel_gram: Float[jnp.ndarray, " D D"],
    noise_variance: ScalarFloat,
    jitter: float = 1e-6,
) -> Float[jnp.ndarray, "D D"]:
    """
    Calculate the covariance matrix for a Gaussian process.

    Parameters
    ----------
    kernel_gram : LinearOperator
        The kernel gram matrix.
    noise_variance : ScalarFloat
        The noise variance.
    jitter : float, optional
        The jitter term to add to the kernel gram matrix, by default 1e-6.
        (This is to ensure the matrix is positive definite.)

    Returns
    -------
    Float[jnp.ndarray, "D D"]
        The covariance matrix for the Gaussian process.

    """
    return kernel_gram + jnp.eye(kernel_gram.shape[0]) * (noise_variance + jitter)
